package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.DDleaf;
import kr.ac.kaist.pomdp.symbolic.perseus.DDnode;
import kr.ac.kaist.pomdp.symbolic.perseus.OP;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;
import kr.ac.kaist.pomdp.utils.SupportList;

import org.apache.log4j.Logger;

/**
 * MaxPlanesBound class for lower bounds in symbolicHSVI algorithm
 * 
 * @see H. S. Sim et al.,
 *      "Symbolic heuristic search value iteration for factored POMDPs,"
 *      AAAI-2008.
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 */
public class MaxPlanesBound implements Serializable {
	private static final long serialVersionUID = -4102777995868543434L;

	private SymbolicProblem p;

	private int horizon; // planning horizon
	private int numNewAlphaCalled = 0;

	public java.util.Vector<Alpha>[] bound; // Bound list
	private int[] lastLPrune; // # of bounds when last pruned
	private int numBackups; // # of backups
	private int[] lastBackupsPrune; // # of backups when last pruned

	// parameters
	private final int PRUNE_PLANE_INCREMENT = 10;
	private final double PRUNE_PLANE_FACTOR = 1.1;
	private final double PRUNE_EPS = 1e-10;
	// private final double TARGET_PRECISION;
	private final int NUM_LOWERBOUND_VECTOR = 1000;

	// Optional Speed-up techniques
	private boolean useExtraPruning = true; // ExtraPruning is powerful for some
											// problems
											// ExtraPruning can prune alpha
											// vector which is not dominated by
											// any other
											// When any node in NodeBunch hold
											// the alpha, we prunes.

	private boolean useCache = true; // Caching is good
	private boolean useSupportList; // Use support list for alpha vector masking
	public boolean useMasking; // Use alpha vector masking

	private SupportList[] supportList;

	private int[][] actPerms;

	/**
	 * MaxPlanesBound constructor
	 * 
	 * @param _p
	 *            input SymbolicProblem
	 * @param _horizon
	 *            planning horizon
	 * @param _bUseMasking
	 *            whether to use masking
	 * @param _actPerms
	 *            action permutations
	 */
	public MaxPlanesBound(SymbolicProblem _p, int _horizon,
			boolean _bUseMasking, int[][] _actPerms) {
		this.horizon = _horizon;
		this.p = _p;
		if (_bUseMasking) {
			this.useSupportList = true;
			this.useMasking = true;
		} else {
			this.useSupportList = false;
			this.useMasking = false;
		}
		this.actPerms = _actPerms;

		bound = new java.util.Vector[horizon];
		lastLPrune = new int[horizon];
		lastBackupsPrune = new int[horizon];
		for (int h = 0; h < horizon; h++) {
			bound[h] = new java.util.Vector(NUM_LOWERBOUND_VECTOR);
			lastLPrune[h] = 0;
			lastBackupsPrune[h] = 0;
		}

		numBackups = 0;

		if (useSupportList) {
			if (useMasking == false) {
				System.err
						.println("useSupportList == true, but useMasking == false");
				System.err.println("setting useMasking true...");
				useMasking = true;
			}

			supportList = new SupportList[horizon];
			for (int h = 0; h < horizon; h++)
				supportList[h] = new SupportList(p);
		}

		// logger
		Loggers.logLower = Logger.getLogger(this.getClass());
	}

	/**
	 * Bound initialization
	 * 
	 * Note that this process is converted to use ADD representations
	 * 
	 * @author Kanghoon Lee
	 * @version 2011.04.08 Apply action permutatons
	 */
	public void initBound(int initHorizon) {
		// HSVI2
		// blind policy value estimate suggested in [Hauskrect, 1997]
		// The initial alpha vector value of each policy "always take action a"
		// is updated in MDP fashion.
		Loggers.logLower.info("create lower bound...");

		// use symmetry
		int[] actionMap = new int[p.nActions];
		for (int ac = 0; ac < p.nActions; ac++) {
			actionMap[ac] = ac;
			for (int p = 0; p < actPerms.length; p++) {
				if (actPerms[p][ac] < actionMap[ac])
					actionMap[ac] = actPerms[p][ac];
			}
		}

		DD[] oldLowers = new DD[p.nActions];
		for (int i = 0; i < p.nActions; i++) {
			oldLowers[i] = DDleaf.myNew(0);
		}

		// HSVI2 blind policy value estimate starts from here.
		DD[] list = new DD[p.actions[0].transFn.length + 2];

		for (int hor = 0; hor < horizon; hor++) {
			DD[] newLowers = new DD[p.nActions];
			DD[] primedV = new DD[p.nActions];

			for (int a = 0; a < p.nActions; a++) {
				if (actionMap[a] != a)
					continue;
				primedV[a] = OP.primeVars(oldLowers[a], p.nVars);
			}

			for (int a = 0; a < p.nActions; a++) {
				// do not add the alpha vector that is symmetrically equivalent
				if (actionMap[a] != a)
					continue;
				list[0] = primedV[a];
				list[1] = DDleaf.myNew(p.discFact);

				for (int i = 0; i < p.actions[a].transFn.length; i++)
					list[i + 2] = p.actions[a].transFn[i];

				newLowers[a] = OP.addMultVarElim(list, p.stateVarsPrimeId);
				DD[] mergeNewAlpha = OP
						.concat(p.actions[a].rewFn, newLowers[a]);
				newLowers[a] = OP.addN(mergeNewAlpha);

				// add alpha vector to the bound list at the horizon h
				Alpha initAlpha = newAlpha(a);
				initAlpha.v = newLowers[a].store();
				if (useMasking) {
					initAlpha.mask = new DD[p.nStateVars];
					for (int j = 0; j < p.nStateVars; j++)
						initAlpha.mask[j] = DD.one;
				}
				addAlpha(hor, initAlpha);
			}

			// for approximating large problems, only initialize to initHorizon
			if (hor >= initHorizon) {
				Loggers.logLower.info("initialize lower bound for horizon #"
						+ hor);
				for (Alpha alpha : bound[hor]) {
					for (int hor2 = hor + 1; hor2 < horizon; hor2++)
						bound[hor2].add(alpha);
				}
				Loggers.logLower.info("alpha vectors between horizon #"
						+ (hor + 1) + " to horizon #" + horizon
						+ " are copied as the horizon # " + hor
						+ "'s alpha vectors");
				break;
			}

			oldLowers = newLowers;
		}
	}

	/**
	 * Initialize node bound. Set the best plane for node.
	 * 
	 * @param cn
	 *            input node
	 */
	public void initNodeBound(Node cn) {
		if (useCache) {
			MaxPlaneData bdata = new MaxPlaneData();
			bdata.bestPlane = null;
			cn.bdata = bdata;
		}

		setPlaneForNode(cn, getBestPlaneValue(cn));

		if (Loggers.logLower.isDebugEnabled()) {
			Alpha bestPlane = getBestPlane(cn);
			Loggers.logLower.debug("opt action = " + bestPlane.a);
			Loggers.logLower.debug("# bounds: " + bound[cn.horizon].size()
					+ ", lbVal: " + cn.lbVal);
		}
	}

	/**
	 * Prune points and planes if the number has grown significantly since the
	 * last check
	 */
	public void mayPrune(int horizon) {
		int LnextNumPrune = Math.max(lastLPrune[horizon]
				+ PRUNE_PLANE_INCREMENT,
				(int) (lastLPrune[horizon] * PRUNE_PLANE_FACTOR));
		int lSize = bound[horizon].size();
		if (lSize > LnextNumPrune) {
			prune(horizon, lSize);

			lastBackupsPrune[horizon] = numBackups;

			if (Loggers.logLower.isDebugEnabled())
				Loggers.logLower.debug("... pruned # planes from " + lSize
						+ " down to " + bound[horizon].size());

		}
	}

	/**
	 * If alpha vector a dominates alpha vector b, return true. Otherwise,
	 * return false.
	 * 
	 * @param a
	 *            input alpha vector
	 * @param b
	 *            input alpha vector
	 * @param ep
	 *            prune eps
	 * @return true or false
	 */
	private boolean dominate(Alpha a, Alpha b, double ep) {
		if (useMasking) {
			return mask_dominates(a.v, b.v, ep, a.mask, b.mask);
		}
		return dominate(a.v, b.v, ep);
	}

	private boolean mask_dominates(DD a, DD b, double ep, DD[] aMask, DD[] bMask) {
		return OP.isSupport(bMask, aMask) && dominate(a, b, ep);
	}

	/**
	 * If DD a dominates DD b, return true. Otherwise, return false. (If max
	 * value of (b-a) is not larger than ep, a dominates b.)
	 * 
	 * @param a
	 *            input DD
	 * @param b
	 *            input DD
	 * @param ep
	 *            prune eps
	 * @return true or false
	 */
	private boolean dominate(DD a, DD b, double ep) {
		if (ep < OP.maxAll(OP.sub(b, a)))
			return false;

		return true;
	}

	/**
	 * Prune dominated alpha vectors
	 * 
	 * @param size
	 *            the number of vectors in current bound list This doesn't
	 *            affect the result of this algorithm.
	 */
	public void prune(int horizon, int size) {
		// if(true)
		// return;
		int numRefCountDeletion = 0;

		// ExtraPruning can prune alpha vector which is not dominated by any
		// other
		// When any node in NodeBunch hold the alpha, we prunes.
		if (useExtraPruning) {
			Iterator<Alpha> it = bound[horizon].iterator();
			while (it.hasNext()) {
				Alpha candidate = it.next();
				if (candidate.backPointers.isEmpty()) {
					it.remove();
					if (useSupportList)
						supportList[horizon].remove(candidate, candidate.mask);
					numRefCountDeletion++;
				}
			}
		}
		if (Loggers.logLower.isDebugEnabled()) {
			Loggers.logLower.debug("(" + numRefCountDeletion
					+ " vectors deleted by refCount)");
		}

		// Prune alphas which are point-wise dominated by others
		for (int i = 0; i < bound[horizon].size(); i++) {
			Alpha candidate = bound[horizon].get(i);
			if (candidate.dirty)
				continue;
			for (int j = i + 1; j < bound[horizon].size(); j++) {
				Alpha target = bound[horizon].get(j);
				if (target.dirty)
					continue;
				if (candidate.numBackupsCreation < lastBackupsPrune[horizon]
						&& target.numBackupsCreation < lastBackupsPrune[horizon]) {
					// candidate and target were compared the last time we
					// pruned and
					// neither dominates the other; leave them both in
					continue;
				} else if (dominate(candidate, target, PRUNE_EPS)) {
					// target is pruned
					target.dirty = true;
					deleteAndForward(horizon, target, candidate);
				} else if (dominate(target, candidate, PRUNE_EPS)) {
					// candidate is pruned
					candidate.dirty = true;
					deleteAndForward(horizon, candidate, target);
				}
			}
		}

		Iterator<Alpha> it = bound[horizon].iterator();
		while (it.hasNext()) {
			Alpha cleanTarget = it.next();
			if (cleanTarget.dirty) {
				it.remove();
			}
		}

		lastLPrune[horizon] = bound[horizon].size();
		lastBackupsPrune[horizon] = numBackups;
	}

	/**
	 * forward backPointers from victim to dominator
	 * 
	 * @param victim
	 * @param dominator
	 */
	public void deleteAndForward(int hor, Alpha victim, Alpha dominator) {
		if (useSupportList) {
			supportList[hor].remove(victim, victim.mask);
		}

		if (useCache && useExtraPruning) {

			// forward backPointers from victim to dominator
			for (int i = 0; i < victim.backPointers.size(); i++) {
				dominator.backPointers.add(victim.backPointers.get(i));
			}
			victim.backPointers.clear();
		}
	}

	/**
	 * update the bound list; get newAlpha for Node cn and add newAlpha to the
	 * bound list
	 * 
	 * @param cn
	 *            input node
	 */
	public void update(Node cn) {
		Alpha newAlpha = getNewLBPlane(cn);
		setPlaneForNode(cn, newAlpha);
		addAlpha(cn.horizon, newAlpha);
	}

	/**
	 * set MaxPlaneData for input node The best plane of input node is changed
	 * to the input alpha vector. New lower bound value is also set.
	 * 
	 * @param cn
	 *            input node
	 * @param newAlpha
	 *            input new alpha
	 */
	public void setPlaneForNode(Node cn, Alpha newAlpha) {

		// new lower bound value
		double newLB = OP.factoredExpectationSparseNoMem(cn.b, newAlpha.v);

		// cache
		if (useCache) {
			if (newLB < cn.lbVal - PRUNE_EPS) {
				return;
			}
			MaxPlaneData bdata = cn.bdata;
			Alpha oldPlane = bdata.bestPlane;

			if (oldPlane != null) {
				oldPlane.backPointers.remove(cn.b);
			}

			// change the best plane of the node to newAlpha
			// add belief state backpointer to newAlpha
			bdata.bestPlane = newAlpha;
			bdata.lastSetPlaneBackups = numBackups;
			newAlpha.backPointers.add(cn.b);
		}
		cn.lbVal = newLB;
	}

	/**
	 * set MaxPlaneData for input node The best plane of input node is changed
	 * to the input alpha vector. New lower bound value is also set.
	 * 
	 * @param cn
	 *            input node
	 * @param newAlpha
	 *            & value input new alpha
	 */
	public void setPlaneForNode(Node cn, ArrayList<Object> newPlaneValue) {

		// new lower bound value
		Alpha newAlpha = (Alpha) newPlaneValue.get(0);
		double newLB = (Double) newPlaneValue.get(1);

		// cache
		if (useCache) {
			if (newLB < cn.lbVal - PRUNE_EPS) {
				return;
			}
			MaxPlaneData bdata = cn.bdata;
			Alpha oldPlane = bdata.bestPlane;

			if (oldPlane != null) {
				oldPlane.backPointers.remove(cn.b);
			}

			// change the best plane of the node to newAlpha
			// add belief state backpointer to newAlpha
			bdata.bestPlane = newAlpha;
			bdata.lastSetPlaneBackups = numBackups;
			newAlpha.backPointers.add(cn.b);
		}
		cn.lbVal = newLB;
	}

	/**
	 * calculate the value of each alpha vector in the bound list at belief b
	 * and return the best plane (having max value at belief b)
	 * 
	 * @param cn
	 *            input node
	 * @return the best plane (having max value at belief b)
	 */
	public Alpha getBestPlane(Node cn) {
		ArrayList<Object> result = getBestPlaneValue(cn);
		return (Alpha) result.get(0);
	}

	/**
	 * calculate the value of each alpha vector in the bound list at belief b
	 * and return the best plane (having max value at belief b)
	 * 
	 * @param cn
	 *            input node
	 * @return the best plane (having max value at belief b)
	 */
	public ArrayList<Object> getBestPlaneValue(Node cn) {
		java.util.Vector<Alpha> planesToCheck = null;

		if (useSupportList) {
			planesToCheck = (java.util.Vector<Alpha>) supportList[cn.horizon]
					.getList(cn.b);
			if (Loggers.logLower.isDebugEnabled())
				Loggers.logLower
						.debug("alpha vector masking reduces # of planes to check from "
								+ bound[cn.horizon].size()
								+ " to "
								+ planesToCheck.size());
		} else {
			planesToCheck = bound[cn.horizon];
		}

		double max = Double.NEGATIVE_INFINITY;
		int iMax = -1; // index of the plane whose value at belief b is max

		// for each plane to check
		for (int i = 0; i < planesToCheck.size(); i++) {
			Alpha cAlpha = planesToCheck.get(i);

			if (useMasking) {
				if (!OP.isSupport(cn.b, cAlpha.mask))
					continue;
			}

			double sum = OP.factoredExpectationSparseNoMem(cn.b, cAlpha.v);
			if (max < sum) {
				max = sum;
				iMax = i;
			}
		}

		// warning: no support list
		if (iMax == -1) {
			planesToCheck = bound[cn.horizon];
			// for each plane to check
			for (int i = 0; i < planesToCheck.size(); i++) {
				Alpha cAlpha = planesToCheck.get(i);

				double sum = OP.factoredExpectationSparseNoMem(cn.b, cAlpha.v);
				if (max < sum) {
					max = sum;
					iMax = i;
				}
			}
		}

		Alpha bestPlane = planesToCheck.get(iMax);
		ArrayList<Object> result = new ArrayList<Object>();
		result.add(bestPlane);
		result.add(max);
		return result;
	}

	/**
	 * get the number of lower bounds
	 * 
	 * @return the number of lower bounds
	 */
	public int getSize() {
		int num = 0;
		for (int h = 0; h < horizon; h++)
			num += bound[h].size();
		return num;
	}

	/**
	 * get the number of lower bounds
	 * 
	 * @return the number of lower bounds
	 */
	public int getSize(int horizon) {
		return bound[horizon].size();
	}

	/**
	 * get the number of lower bounds for each horizon
	 * 
	 * @return the number of lower bounds for each horizon
	 */
	public String getSizeString() {
		String result = "[";
		for (int h = horizon - 1; h >= 1; h--)
			result += bound[h].size() + ", ";
		result += bound[0].size() + "]";
		return result;
	}

	/**
	 * add input alpha vector to this bound list and increase backup number by 1
	 * 
	 * @param hor
	 *            horizon
	 * @param plane
	 *            input alpha vector
	 */
	public void addAlpha(int hor, Alpha plane) {
		// backup # when this alpha vector is created
		plane.numBackupsCreation = numBackups;

		if (Loggers.logLower.isDebugEnabled()) {
			// log.debug("add alpha");
			// plane.v.display();
		}

		if (useSupportList) {
			// add new alpha vector to support list
			supportList[hor].add(plane, plane.mask);
		}

		if (plane.primedV == null)
			plane.primedV = OP.primeVars(plane.v, p.nVars);

		bound[hor].add(plane);
		numBackups++;
	}

	/**
	 * return the array of alpha vectors
	 * 
	 * @return array of alpha vectors [horizon][index of alpha vector]
	 */
	public Vector<Alpha>[] getAlphas() {
		return bound;
	}

	/**
	 * return new alpha vector corresponding to action a
	 * 
	 * @param a
	 *            input action a
	 * @return new alpha vector
	 */
	public Alpha newAlpha(int a) {
		Alpha res = new Alpha();
		res.dirty = false;
		res.a = a;
		return res;
	}

	/**
	 * return new alpha vector and new value at belief point of input node cn.
	 * 
	 * @param cn
	 *            input node
	 * @return New alpha vector
	 */
	public Alpha getNewLBPlane(Node cn) {
		numNewAlphaCalled++;
		if (Loggers.logLower.isDebugEnabled())
			Loggers.logLower.debug("# alpha = " + numNewAlphaCalled);

		// -------------------------------------------------------------
		// i) find best action and observation strategy at belState
		// -------------------------------------------------------------

		// for each action,
		double bestValue = Double.NEGATIVE_INFINITY;
		int bestActId = -1;
		for (int actId = 0; actId < p.nActions; actId++) {
			// calculate next value for each alpha vector
			double sum_obsValues = _nextValue(cn.Q[actId].outcomes);
			double actValue = cn.Q[actId].immediateReward;
			actValue += (sum_obsValues * p.discFact);
			cn.Q[actId].lbVal = actValue;

			// adjust bestValue
			if (actValue > bestValue) {
				bestValue = actValue;
				bestActId = actId;
			}
		}

		if (Loggers.logLower.isDebugEnabled())
			Loggers.logLower.debug("best action = " + bestActId);

		Alpha newAlpha = getNewAlpha(cn, bestActId);
		return newAlpha;
	}

	/**
	 * return the maximum edge value
	 * 
	 * @param dd
	 *            input edge
	 * @return maximum next value of the dd
	 */
	private double _nextValue(EdgeDd dd) {

		double sum = 0;
		// it's a leaf
		if (dd.getVar() == 0) {
			Edge childEdge = dd.getEdge();
			if (childEdge.obsProb == 0) {
				return 0;
			}
			Node childNode = childEdge.nextNode;

			EdgeDdLeaf ddLeaf = (EdgeDdLeaf) dd;

			ArrayList<Object> bestPlaneValue = getBestPlaneValue(childNode);
			Alpha bestPlane = (Alpha) bestPlaneValue.get(0);
			double bestValue = (Double) bestPlaneValue.get(1);

			ddLeaf.obsStrat = bestPlane;

			double maxNextValue = childEdge.mult * childEdge.obsProb
					* bestValue;
			return maxNextValue;
		} else {
			EdgeDd[] children = dd.getChildren();
			for (int i = 0; i < children.length; i++) {
				sum += _nextValue(children[i]);
			}
		}

		return sum;
	}

	/**
	 * construct corresponding new alpha vector
	 * 
	 * @param cn
	 *            current node
	 * @param bestActId
	 * @return new alpha vector
	 */
	public Alpha getNewAlpha(Node cn, int bestActId) {
		DD[] b = cn.b;
		QEntry qa = cn.Q[bestActId];

		// -------------------------------------------------------------
		// ii) construct corresponding alpha vector
		// -------------------------------------------------------------
		DD newAlpha = DD.zero;
		DD[] mask = null;

		Alpha defaultAlpha = getBestPlane(cn);
		DD nextValFn = getPrimeValueDD(qa.outcomesDd, qa.outcomes, defaultAlpha);

		java.util.Vector dds = new java.util.Vector(p.nVars * 2);
		Common.addList(dds, DDleaf.myNew(p.discFact));
		if (useMasking) {
			mask = new DD[p.nStateVars];
			for (int s = 0; s < p.nStateVars; s++)
				mask[s] = OP.mask(b[s]);
			Common.addList(dds, mask);
		}
		Common.addList(dds, p.actions[bestActId].transFn);
		Common.addList(dds, p.actions[bestActId].obsFn);
		Common.addList(dds, nextValFn);

		newAlpha = OP.addMultVarElim(Common.convert2array(dds),
				p.stateObsVarsIds);
		if (useMasking) {
			newAlpha = OP.add(
					maskAlphaWithMask(p.actions[bestActId].rewFn, mask),
					newAlpha);
		} else {
			newAlpha = OP.add(p.actions[bestActId].rewFn, newAlpha);
		}

		Alpha bestAlpha = newAlpha(bestActId);
		bestAlpha.v = newAlpha;
		if (useMasking)
			bestAlpha.mask = mask;
		return bestAlpha;
	}

	/**
	 * search next values using observation probability and
	 * 
	 * @param ospProbs
	 * @param edgeDd
	 * @param defaultAlpha
	 *            the default alpha vector. This is used when the observation
	 *            probability is zero
	 * @return
	 */
	private DD getPrimeValueDD(DD ospProbs, EdgeDd edgeDd, Alpha defaultAlpha) {
		int ospHead = ospProbs.getVar();

		// ospProbs reached terminal or state var
		if (ospHead == 0 || ospHead < p.obsVarsPrimeId[0]) {
			if (edgeDd.getEdge().obsProb == 0)
				return defaultAlpha.primedV;

			// inv perm
			Edge edge = edgeDd.getEdge();
			DD origAl = ((EdgeDdLeaf) edgeDd).obsStrat.primedV;

			DD alphaAO = OP.swapVars(origAl, edge.invPerm);
			if (useMasking) {
				alphaAO = maskAlphaWithBelief(alphaAO,
						edgeDd.getEdge().nextNode.b);
			}
			return alphaAO;
		} else {
			// recursion
			DD[] children = ospProbs.getChildren();
			DD[] newchildren = new DD[children.length];
			for (int i = 0; i < children.length; i++) {
				newchildren[i] = getPrimeValueDD(children[i],
						edgeDd.getChildren()[i], defaultAlpha);
			}

			return DDnode.myNew(ospHead, newchildren);
		}
	}

	/**
	 * return masked value (value is 0 in the state where belief is zero)
	 * 
	 * @param value
	 * @param mask
	 * @return
	 */
	private DD maskAlphaWithMask(DD value, DD[] mask) {
		java.util.Vector dds = new java.util.Vector();
		Common.addList(dds, value);
		Common.addList(dds, mask);
		DD maskedValue = OP.multN(Common.convert2array(dds));
		return maskedValue;
	}

	/**
	 * return masked value (value is 0 in the state where belief is zero)
	 * 
	 * @param value
	 * @param b
	 * @return
	 */
	private DD maskAlphaWithBelief(DD value, DD[] b) {
		// mask of belief b
		DD[] mask = new DD[p.nStateVars];
		for (int s = 0; s < p.nStateVars; s++)
			mask[s] = OP.mask(OP.primeVars(b[s], p.nVars));

		java.util.Vector dds = new java.util.Vector();
		Common.addList(dds, value);
		Common.addList(dds, mask);
		DD maskedValue = OP.multN(Common.convert2array(dds));
		return maskedValue;
	}
}
