package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.DDleaf;
import kr.ac.kaist.pomdp.symbolic.perseus.OP;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;
import kr.ac.kaist.pomdp.utils.Util;

import org.apache.log4j.Logger;

/**
 * SawtoothBound
 * for upper bounds in SymbolicHsvi algorithm
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SawtoothBound implements Serializable
{
	private static final long serialVersionUID = -7668057062626974270L;

	private int horizon; // planning horizon
	public java.util.Vector<BoundPair>[] bound; // bound pair for each horizon
	private int numBackups;
	private SymbolicProblem p;

	private int[] lastUPrune; // # of bounds when last pruned
	private int[] lastBackupsPrune; // # of backups when last pruned
	public DD[] cornerPts; // Belief simplex corner point values

	// inner parameters
	private final int PRUNE_PTS_INCREMENT = 10;
	private final int PRUNE_PTS_FACTOR = 2;
	private final int NUM_UPPERBOUND_PTS = 1000; // maximum number of upper bounds
	private final double PRUNE_EPS = 1e-10;
	private final double CORNER_EPS = 1e-6;
	//	private final double BELLMAN_EPS = 1e-6;

	private int[][] varPerms;
	private int[][] actPerms;
	
	/**
	 * SawtoothBound constructor
	 * 
	 * @param p
	 * 			input SymbolicProblem
	 * @param _horizon
	 * 			planning horizon
	 * @param _bUseMasking
	 * 			whether to use masking
	 * @param _varPerms
	 * 			variable permutations
	 * @param _actPerms
	 * 			action permutations
	 */
	public SawtoothBound(SymbolicProblem _p, int _horizon, int[][] _varPerms, int[][] _actPerms)
	{
		this.horizon = _horizon;
		this.p = _p;
		this.varPerms = _varPerms;
		this.actPerms = _actPerms;
				
		bound = new java.util.Vector[horizon];
		lastUPrune = new int[horizon];
		lastBackupsPrune = new int[horizon];
		for (int h = 0; h < horizon; h++)
		{
			bound[h] = new java.util.Vector(NUM_UPPERBOUND_PTS);
			lastUPrune[h] = 0;
			lastBackupsPrune[h] = 0;
		}

		numBackups = 0;
		
		// logger
		Loggers.logUpper = Logger.getLogger(this.getClass());
	}

	/**
	 * Prune bounds if the number has grown significantly since the last check
	 * 
	 * @param hor
	 * 			horizon
	 */
	public void mayPrune(int hor)
	{
		int UnextNumPrune =
				Math.max(lastUPrune[hor] + PRUNE_PTS_INCREMENT,
						(int) (lastUPrune[hor] * PRUNE_PTS_FACTOR));
		int uSize = bound[hor].size();

		if (uSize > UnextNumPrune)
		{
			prune(hor, uSize);
			lastBackupsPrune[hor] = numBackups;

			if (Loggers.logUpper.isDebugEnabled())
				Loggers.logUpper.debug("... pruned # points from " + uSize + " down to " + bound[hor].size());
		}
	}

	/**
	 * Prune dominated bounds
	 * 
	 * @param hor
	 * 			horizon
	 * @param size
	 * 			the number of the upper bound list
	 * 			This doesn't affect the result of this algorithm. 
	 */
	public void prune(int hor, int size)
	{

		// Is it a prune time?
		// before pruning, projectedV must be calculated to test bound domination
		for (int i = 0; i < bound[hor].size(); i++)
		{
			BoundPair bp = bound[hor].get(i);
			bp.projectedV = OP.factoredExpectationSparseNoMem(bp.b, cornerPts[hor]);
		}

		for (int i = 0; i < bound[hor].size(); i++)
		{
			BoundPair candidate = bound[hor].get(i);
			if (candidate.dirty)
				continue;

			for (int j = i + 1; j < bound[hor].size(); j++)
			{
				BoundPair target = bound[hor].get(j);
				if (candidate.numBackupsCreation < lastBackupsPrune[hor]
						&& target.numBackupsCreation < lastBackupsPrune[hor])
				{
					// candidate and target were compared the last time we pruned and
					// neither dominates the other; leave them both in
					continue;
				}
				else if (dominate(candidate, target))
				{
					target.dirty = true;
				}
				else if (dominate(target, candidate))
				{
					candidate.dirty = true;
				}
			}
		}

		// remove dirty bounds
		Iterator<BoundPair> it = bound[hor].iterator();
		while (it.hasNext())
		{
			BoundPair cleanTarget = it.next();
			if (cleanTarget.dirty)
			{
				it.remove();
			}
		}

		lastUPrune[hor] = bound[hor].size();
	}

	/**
	 * return true iff candidate dominates target.
	 * NOTE: projectedV variables must be set properly before dominate() is called!
	 * 
	 * @param candidate
	 * 			input BoundPair
	 * @param target
	 * 			input BoundPair
	 * @return true or false
	 */
	private boolean dominate(BoundPair candidate, BoundPair target)
	{
		double cValue = getBVValue(target.b, candidate, target.projectedV, candidate.projectedV);
		double tValue = target.V;
		return (cValue < tValue + PRUNE_EPS);
	}

	/**
	 * returns the upper bound that the BoundPair cpair induces on b.
	 * 
	 * @param b
	 * 			input belief point
	 * @param cpair
	 * 			input BoundPair
	 * @param innerPtsB
	 * @param innerPtsC
	 * @return upper bound on belief point b
	 */
	private double getBVValue(DD[] b, BoundPair cpair, double innerPtsB, double innerPtsC)
	{

		if (innerPtsC <= cpair.V)
		{
			// cpair.b does not provide a useful bound because it actually lies above the plane defined by cornerPts.
			// Ideally, we would prune cpair.b here.
			return Double.POSITIVE_INFINITY;
		}

		// For each variable
		double minRatio = 1; // minimum of belief proportions

		// NOTE: factored belief approximation technique is used.
		// We can compute the upper bound value as the product of the minimum of belief marginal ratios
		// without explicitly enumerating all the states
		for (int s = 0; s < p.nStateVars; s++)
		{

			double ratio = Common.div(b[s], cpair.b[s]);
			if (ratio < CORNER_EPS)
				return Double.POSITIVE_INFINITY;
			if (ratio == Double.POSITIVE_INFINITY)
				return Double.POSITIVE_INFINITY;

			minRatio *= ratio;
		}

		double result = innerPtsB + minRatio * (cpair.V - innerPtsC);
		return result;
	}

	/**
	 * get upper bound value of node cn
	 *  
	 * @param cn
	 * 			input node
	 * @return upper bound value
	 */
	public double getValue(Node cn)
	{
		double innerPtsB = OP.factoredExpectationSparseNoMem(cn.b, cornerPts[cn.horizon]);
		double minValue = innerPtsB;

		java.util.Vector<BoundPair> ptsCandidate;
		ptsCandidate = bound[cn.horizon];

		for (int i = 0; i < ptsCandidate.size(); i++)
		{
			BoundPair bp = ptsCandidate.get(i);
			double innerPtsC = OP.factoredExpectationSparseNoMem(bp.b, cornerPts[cn.horizon]);

			double BVvalue = getBVValue(cn.b, bp, innerPtsB, innerPtsC);
			if (minValue > BVvalue)
			{
				minValue = BVvalue;
			}
		}
		return minValue;
	}

	/**
	 * Belief point exploration
	 * update upperbound and get the optimal action aStar in r
	 * 
	 * @param cn
	 * 			input node
	 * @param r
	 * 			the optimal action
	 */
	public void update(Node cn, ArrayList r)
	{
		double newUBVal = getNewUBValue(cn, r);
		setUBForNode(cn, newUBVal, true);
	}

	private void setUBForNode(Node cn, double newUB, boolean addBV)
	{
		cn.ubVal = newUB;

		if (addBV)
		{
			// add all belief permutations
			int[][] varMapping = new int[2][p.nStateVars];
			for(int i=0; i<p.nStateVars; i++) {
				varMapping[0][i] = i+1;
			}
			for(int perm=0; perm<varPerms.length; perm++) {
				DD[] newB = new DD[p.nStateVars];
				varMapping[1] = varPerms[perm].clone();
				for(int i=0; i<cn.b.length; i++) {
					newB[i] = OP.swapVars(cn.b[i], varMapping);
				}
				BoundPair pt = new BoundPair(newB, newUB);
				addPoint(cn.horizon, pt);
			}
		}
	}

	private double getNewUBValue(Node cn, ArrayList r)
	{
		if (cn.Q[0].ubVal == Double.NEGATIVE_INFINITY)
			return getNewUBValueSimple(cn, r);
		else
			return getNewUBValueCache(cn, r);
	}

	/**
	 * update upperbound and get the optimal action aStar in r
	 * upper bound value (expected reward) is already updated.
	 * 
	 * @param cn
	 * 			input node
	 * @param r
	 * 			the optimal action
	 * @return new upper bound value
	 */
	private double getNewUBValueCache(Node cn, ArrayList r)
	{

		// cache upper bound for each action
		double[] cachedUpperBound = new double[p.nActions];
		// remember which Q functions we have updated on this call
		boolean[] updatedAction = new boolean[p.nActions];

		for (int a = 0; a < p.nActions; a++)
		{
			cachedUpperBound[a] = cn.Q[a].ubVal;
			updatedAction[a] = false;
		}

		double val;
		// action whose expected reward is maximum
		int maxUBAction = Util.argmax(cachedUpperBound);
		while (true)
		{

			// start getNewUBValueQ: do the backup for the best Q
			// upper bound on long-term reward for taking action maxUBAction
			val = getNewUBValueQ(cn, maxUBAction);
			cachedUpperBound[maxUBAction] = val;
			updatedAction[maxUBAction] = true;

			// the best action may have changed after updating Q
			maxUBAction = Util.argmax(cachedUpperBound);

			// if the best action after the update is one that we have already updated, we're done
			if (updatedAction[maxUBAction])
				break;
		}

		double maxVal = cachedUpperBound[maxUBAction];
		r.add(maxUBAction);

		return maxVal;
	}

	/**
	 * update upperbound and get the optimal action aStar in r
	 * upper bound value (expected reward) is not updated. (NEGATIVE_INFINITY)
	 * 
	 * @param cn
	 * 			input node
	 * @param r
	 * 			the optimal action
	 * @return new upper bound value
	 */
	private double getNewUBValueSimple(Node cn, ArrayList r)
	{

		double maxQba = Double.NEGATIVE_INFINITY;
		int imaxQba = -1; // best action
		double val;

		// for each action, calculate Q
		for (int a = 0; a < p.nActions; a++)
		{

			// find max Qvalue and corresponding best action
			val = getNewUBValueQ(cn, a);
			if (val > maxQba)
			{
				maxQba = val;
				imaxQba = a;
			}
		}

		r.add(imaxQba);
		return maxQba;
	}

	/**
	 * upper bound on long-term reward for taking action a
	 * @param cn node
	 * @param a action
	 * @return one look-ahead q-value
	 */
	private double getNewUBValueQ(Node cn, int a)
	{
		double val = 0;

		QEntry Qa = cn.Q[a];

		// for each observation, (symbolic observation version)
		val = _getNewUBValueQ(Qa.outcomes);

		// update expected reward
		val = Qa.immediateReward + p.discFact * val;
		Qa.ubVal = val;

		return val;
	}

	private double _getNewUBValueQ(EdgeDd dd)
	{
		// it's a leaf
		double val = 0.0;
		if (dd.getVar() == 0)
		{
			Edge e = dd.getEdge();
			if (e.obsProb > 0)
			{
				val += e.mult * e.obsProb * getValue(e.nextNode);
			}
		}
		else
		{
			EdgeDd[] children = dd.getChildren();
			for (int i = 0; i < children.length; i++)
			{
				val += _getNewUBValueQ(children[i]);
			}
		}

		return val;
	}

	/**
	 * add BoundPair pt to the upper bound list
	 * 
	 * @param pt
	 * 			input BoundPair
	 */
	private void addPoint(int horizon, BoundPair pt)
	{
		pt.numBackupsCreation = numBackups;
		bound[horizon].add(pt);

		if (Loggers.logUpper.isDebugEnabled())
		{
			Loggers.logUpper.debug("(newly added UB)" + pt.V);
		}
		numBackups++;
	}

	/**
	 * return the number of upper bounds
	 * 
	 * @return the number of upper bounds
	 */
	public int getSize()
	{
		int num = 0;
		for (int h = 0; h < horizon; h++)
			num += bound[h].size();
		return num;
	}

	/**
	 * return the number of upper bounds
	 * 
	 * @return the number of upper bounds
	 */
	public int getSize(int horizon)
	{
		return bound[horizon].size();
	}

	public String getSizeString()
	{
		String result = "[";
		for (int h = horizon - 1; h >= 1; h--)
			result += bound[h].size() + ", ";
		result += bound[0].size() + "]";
		return result;
	}

	/**
	 * Compute initial upper bound method as baseline using fully observable MDP value iteration
	 * 
	 * @return MDP value iteration result
	 * 			(set of all belief simplex corner points for each state and their corresponding upper bound value)
	 * 
	 * @author Dongho Kim
	 * @version 2010.12.15 
	 * BUG FIX: 1. change error precision in value iteration to use BELLMAN_EPS
	 * 			//2. initialize alpha vectors by Vmax to guarantee the upper bound properties
	 * 			2. ZMDP package uses zero-vectors as initial values
	 * 
	 * @author Kanghoon Lee
	 * @version 2011.04.08 
	 * Apply action permutations
	 * 
	 */
	public void runVmdp(int initHorizon)
	{
		DD currentValue = DDleaf.myNew(0);
		DD nextValue;

		// find the actions that are symmetrically same 
		int[] actionMap = new int[p.nActions];
		for(int ac=0; ac<p.nActions; ac++) {
			actionMap[ac] = ac;
			for(int p=0; p<actPerms.length; p++) {
				if(actPerms[p][ac] < actionMap[ac])
					actionMap[ac] = actPerms[p][ac];
			}
		}
		
		DD[][] list = new DD[p.nActions][p.nStateVars + 1];
		for (int a = 0; a < p.nActions; a++) {
			for (int s = 0; s < p.nStateVars; s++)
				list[a][s] = p.actions[a].transFn[s];
		}
		DD disc = DDleaf.myNew(p.discFact);
		
		// start MDP value iteration
		for (int hor = 0; hor < horizon; hor++)
		{
			DD[] newValue = new DD[p.nActions];
			DD primedV = OP.primeVars(currentValue, p.nVars);
			DD discValue = OP.mult(primedV, disc);

			// for each action
			for (int a = 0; a < p.nActions; a++)
			{
				if(actionMap[a] == a) {
					// backup
					list[a][p.nStateVars] = discValue;
					newValue[a] = OP.addMultVarElim(list[a], p.stateVarsPrimeId);
					DD[] mergeNewAlpha = OP.concat(p.actions[a].rewFn, newValue[a]);
					newValue[a] = OP.addN(mergeNewAlpha);
				} else {
					// use the value of symmetrically equivalent
					newValue[a] = newValue[actionMap[a]];
					int[][] varMapping = new int[2][p.nStateVars*2];
					for(int pm=0; pm<actPerms.length; pm++) {
						if(actPerms[pm][a] == actionMap[a]) {
							for(int stVar=0; stVar<p.nStateVars; stVar++) {
								varMapping[1][stVar] = stVar+1;
								varMapping[1][stVar+p.nStateVars] = stVar+1+p.nVars;
								varMapping[0][stVar] = varPerms[pm][stVar];
								varMapping[0][stVar+p.nStateVars] = varPerms[pm][stVar]+p.nVars;
							}
							break;
						}
					}
					newValue[a] = OP.swapVars(newValue[a], varMapping);
				}

			}

			// find maximum nextValue
			nextValue = newValue[0];
			for (int a = 1; a < p.nActions; a++) {
				nextValue = OP.max(nextValue, newValue[a]);
			}
			
			cornerPts[hor] = nextValue.store();
			currentValue = nextValue;

			// for approximating large problems, only initialize to initHorizon
			if(hor >= initHorizon) {
				Loggers.logUpper.info("initialize upper bound for horizon #" + hor);
				for(int hor2=hor+1; hor2<horizon; hor2++)
					cornerPts[hor2] = nextValue.store();
				Loggers.logUpper.info("alpha vectors between horizon #" + (hor+1) + " to horizon #" + 
						horizon + " are copied as the horizon # " + hor + "'s alpha vectors");
				break;
			}
		}
	}

	/**
	 * Initialize upper bound
	 */
	public void initBound(int initHorizon)
	{
		// upperbound - HSVI1 (VMDP)
		Loggers.logUpper.info("create upper bound...");
		cornerPts = new DD[horizon];
		runVmdp(initHorizon);
	}

	/**
	 * initialize node upper bound
	 * 
	 * @param cn
	 * 			input node
	 */
	public void initNodeBound(Node cn)
	{
		cn.ubVal = getValue(cn);
	}
}
