package kr.ac.kaist.pomdp.symbolic.hsvi.ippc2011;

import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.MySet;
import kr.ac.kaist.pomdp.symbolic.perseus.OP;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;
import kr.ac.kaist.pomdp.utils.Util;

import no.uib.cipr.matrix.Matrix;

import org.apache.log4j.Logger;

/**
 * Symbolic Heuristic Search Value Iteration for factored POMDPs
 * 
 * (symbolic observation version)
 * 
 * @see T. Smith and R. Simmons, "Heuristic search value iteration for POMDPs,"
 *      UAI-2004.
 * @see T. Smith and R. Simmons,
 *      "Point-based POMDP algorithms: Improved analysis and implementation,"
 *      UAI-2005.
 * @see H. S. Sim et al.,
 *      "Symbolic heuristic search value iteration for factored POMDPs,"
 *      AAAI-2008.
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SymbolicHsviIPPC2011 implements Serializable {
	private static final long serialVersionUID = -7489176876878725980L;

	public int horizon; // planning horizon (finite)

	public SawtoothBound upperBound; // upper bound
	public MaxPlanesBound lowerBound; // lower bound

	private SymbolicProblem p; // symbolic problem
	private double epsilon; // GOAL WIDTH

	private int saveDepth; // Deepest depth of node ever traversed
	public static double SAME_EPS = 1e-10;
	private int nodeHits;

	private HashMap<Node, Node> nodeBunch; // Node Bunch

	public DD[] initBelState; // factored representation of initial belief
	public DD initBelState_orig; // initial belief
	public ArrayList HSresult; // Global variable to store Heuristic Search
								// result
	private Node rootNode; // root node of search tree
	private int iter;

	private String policyFilename; // policy filename to save

	private int[][] varPerms; // permutations for state, obs variables
	private int[][] actPerms; // permutations for actions

	private int initHorizon; // for approximating, only initialize to
								// initHorizon
	private boolean bUseMasking; // whether to use masking
	private boolean bUseSymmetry; // whether to use symmetry

	private boolean bPeSim; // whether to simulate using current policy
							// periodically
	private int nPeSimNum; // # simulations

	/**
	 * SymbolicHsviSymObs constructor
	 * 
	 * @param _pomdp
	 *            input SymbolicProblem
	 * @param _ep
	 *            Goal width
	 * @param _initHoizon
	 *            when initializing, the number of initializing steps
	 * @param _bUseMaksing
	 *            whether to use masking
	 * @param _bUseSymmetry
	 *            whether to use symmetry
	 * @param _policyFilename
	 *            policy filename
	 * @param _permFilename
	 *            permutation filename
	 * @param _bPeSim
	 *            whether to simulate
	 * @param _nPeSimNum
	 *            the number of simulations
	 */
	public SymbolicHsviIPPC2011(SymbolicProblem _pomdp, double _ep,
			int _initHorizon, boolean _bUseMasking, boolean _bUseSymmetry,
			String _policyFilename, String _permFilename, boolean _bPeSim,
			int _nPeSimNum) {
		this.p = _pomdp;
		this.epsilon = _ep;
		this.initHorizon = Math.min(_initHorizon, p.horizon);
		this.policyFilename = _policyFilename;
		this.horizon = p.horizon;
		// read the permuations
		this.bUseSymmetry = _bUseSymmetry;
		ArrayList permList;
		if (!bUseSymmetry) {
			this.actPerms = null;
			this.varPerms = null;
			permList = readPermutations("");
		} else {
			permList = readPermutations(_permFilename);
		}
		this.varPerms = (int[][]) permList.get(0);
		this.actPerms = (int[][]) permList.get(1);
		p.setPerms(varPerms, actPerms);
		this.bUseMasking = _bUseMasking;

		// create lower, upper bounds
		lowerBound = new MaxPlanesBound(p, horizon, bUseMasking, actPerms);
		upperBound = new SawtoothBound(p, horizon, varPerms, actPerms);

		nodeBunch = new HashMap();
		nodeHits = 0;

		// Factorize initBelState
		initBelState_orig = p.initialBelState;
		DD[] temp = new DD[p.nStateVars];
		for (int varId = 0; varId < p.nStateVars; varId++) {
			int[] vars = Util.elimOneVar(p.stateVarsId, varId);
			vars = MySet.remove(vars, 0);
			temp[varId] = OP.addMultVarElim(p.initialBelState, vars);
		}
		initBelState = temp;
		saveDepth = 1;
		HSresult = new ArrayList();
	}

	/**
	 * Run symbolicHSVI (AnytimeHSVI)
	 * 
	 * @param precisionRatio
	 *            precision ratio
	 * @throws Exception
	 */
	public void run(double precisionRatio) throws Exception {
		double err = Double.POSITIVE_INFINITY;
		double precision;

		Loggers.logSymHsviDdTreeFiniteSymmetry = Logger.getLogger(this
				.getClass());
		Loggers.logLower = Logger.getLogger(this.lowerBound.getClass());
		Loggers.logUpper = Logger.getLogger(this.upperBound.getClass());

		Loggers.logSymHsviDdTreeFiniteSymmetry.info("horizon: " + horizon);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("use masking: "
				+ bUseMasking);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("use symmetry: "
				+ bUseSymmetry);

		initBound();

		iter = 0;

		// root Node initialize
		rootNode = new Node(initBelState, p.stateVarsId, horizon - 1);
		ArrayList bPerm = sortStVars(rootNode.b);
		rootNode.b = (DD[]) bPerm.get(0);

		lowerBound.initNodeBound(rootNode);
		upperBound.initNodeBound(rootNode);
		nodeBunch.put(rootNode, rootNode);

		savePolicy();

		double initialUB = rootNode.ubVal;
		double initialLB = rootNode.lbVal;
		double initialWidth = initialUB - initialLB;
		err = initialWidth;

		Loggers.logSymHsviDdTreeFiniteSymmetry.info("initial bound ["
				+ initialLB + " ... " + initialUB + "]");

		long startTime = System.currentTimeMillis();
		int startIter = iter;
		// Run iterations
		while (err > epsilon) {
			iter++;

			// Set target precision
			precision = (rootNode.ubVal - rootNode.lbVal) * precisionRatio;
			printOpenEpisode(precision);

			// Heuristic-search Exploration
			// AnytimeHSVI
			explore(rootNode, precision, 0);

			// print a progress update
			err = rootNode.ubVal - rootNode.lbVal;
			long elapsedTime = System.currentTimeMillis() - startTime;
			int elapsedIter = iter - startIter;
			if (elapsedTime > 100000 || elapsedIter > 20) {
				printCloseEpisode(err);
				startTime = System.currentTimeMillis();
				startIter = iter;

				// simulate
				if (bPeSim)
					simulate(nPeSimNum);
			}

		}

		// After iterations, prune last
		for (int h = 0; h < horizon; h++) {
			lowerBound.prune(h, lowerBound.getSize(h));
			upperBound.prune(h, upperBound.getSize(h));
		}
		savePolicy();
		simulate(100);

		// final results
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("err: " + err + ", "
				+ rootNode.lbVal + "..." + rootNode.ubVal);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of iteration: " + iter);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of lowerbound: "
				+ lowerBound.getSize() + "\t" + lowerBound.getSizeString());
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of upperbound: "
				+ upperBound.getSize() + "\t" + upperBound.getSizeString());
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of created node: "
				+ nodeBunch.size());
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of node hits: "
				+ nodeHits);
	}

	/**
	 * simulate using the current policy
	 * 
	 * @param nRuns
	 *            the number of simulation episodes
	 */
	private void simulate(int nRuns) {
		int nSteps = p.horizon;
		if (nSteps == Integer.MAX_VALUE)
			nSteps = 40;

		Loggers.logSymHsviDdTreeFiniteSymmetry.info("Simulation start ("
				+ nRuns + " times, " + nSteps + " steps)");

		java.util.ArrayList alphas = getAlphas();
		Matrix result;
		if (alphas.size() == 2) {
			// not use masking
			result = p
					.sampleRewardsFinite((java.util.Vector[]) alphas.get(0),
							(java.util.Vector[]) alphas.get(1), nRuns, nSteps,
							0, false);
		} else {
			// use masking
			result = p
					.sampleRewardsFiniteMask(
							(java.util.Vector[]) alphas.get(0),
							(java.util.Vector[]) alphas.get(1),
							(java.util.Vector[]) alphas.get(2), nRuns, nSteps,
							0, false);
		}

		// print reward sampling results
		double sum = 0;
		int sum_step = 0;

		for (int i = 0; i < result.numRows(); i++) {
			sum += result.get(i, 0);
			sum_step += result.get(i, 1);
		}

		double avg = sum / result.numRows();
		double var = 0;
		for (int i = 0; i < result.numRows(); i++)
			var += Math.pow((result.get(i, 0) - avg), 2);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("average reward = " + avg
				+ "  (std: " + Math.sqrt(var / (result.numRows() - 1)) + ")");
	}

	/**
	 * save only the current policy to file
	 */
	private void savePolicy() {
		try {
			if (policyFilename != null) {
				Loggers.logSymHsviDdTreeFiniteSymmetry
						.info("Writing SymHsviDdTreeFiniteSymmetry Policy... ");
				java.util.ArrayList alphas = getAlphas();
				ObjectOutputStream alphasOut = new ObjectOutputStream(
						new FileOutputStream(policyFilename));
				alphasOut.writeObject(alphas);
				alphasOut.close();
			}
		} catch (Exception e) {
			System.out.println(e);
		}
	}

	/**
	 * Print starting
	 */
	private void printOpenEpisode(double precision) {

		if (Loggers.logSymHsviDdTreeFiniteSymmetry.isDebugEnabled()) {
			Loggers.logSymHsviDdTreeFiniteSymmetry.debug("-*- doTrial: trial "
					+ iter + "\nub =" + rootNode.ubVal + ", lb ="
					+ rootNode.lbVal + ", ");
		} else if (iter == 1 || iter % 10 == 0) {
			Loggers.logSymHsviDdTreeFiniteSymmetry.info("-*- doTrial: trial "
					+ iter);
		}
		if (Loggers.logSymHsviDdTreeFiniteSymmetry.isDebugEnabled()) {
			Loggers.logSymHsviDdTreeFiniteSymmetry.debug("target precision: "
					+ precision);
		}
	}

	/**
	 * Print a progress update every 10 seconds
	 */
	private void printCloseEpisode(double err) {
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("err: " + err + ", "
				+ rootNode.lbVal + "..." + rootNode.ubVal);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of lowerbound: "
				+ lowerBound.getSize());
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("# of upperbound: "
				+ upperBound.getSize());
	}

	/**
	 * Explore recursively follows a single path down the search tree until
	 * satisfying a termination condition based on the width of the bounds
	 * interval. It then performs a series of updates on its way back up to the
	 * initial belief.
	 * 
	 * @param cn
	 *            exploring node
	 * @param ep
	 *            Goal width
	 * @param depth
	 *            depth of the node
	 */
	private void explore(Node cn, double ep, int depth) throws Exception {
		if (saveDepth < depth)
			saveDepth = depth;

		// Excess Uncertainty
		double excessUnc = cn.ubVal - cn.lbVal - ep
				* Math.pow(p.discFact, -depth);
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("exploring depth = "
				+ depth + ",  excessUnc = " + excessUnc);

		// Terminating condition
		if (excessUnc <= 0 || depth >= horizon) {
			// Terminate
			if (Loggers.logSymHsviDdTreeFiniteSymmetry.isDebugEnabled()) {
				Loggers.logSymHsviDdTreeFiniteSymmetry
						.debug("  trialRecurse: depth=" + depth + ", ["
								+ cn.lbVal + " .. " + cn.ubVal
								+ "], excessUnc = " + excessUnc
								+ " (terminating)");
			}
			return;
		}

		// Expand node. setting upper and lower values of children
		expand(cn);

		// First bound update
		// Get an action aStar and observation oStar according to the forward
		// exploration heuristics
		HSresult.clear();
		update(cn, ep, depth, HSresult);

		Node maxExcessNode = (Node) HSresult.get(1);

		// explore successor node
		explore(maxExcessNode, ep, depth + 1);
		HSresult.clear();

		// Second bound update
		update(cn, ep, depth, HSresult);
	}

	/**
	 * for "update" function Get aStar and oStar which make the child in depth+1
	 * to have most contribution to excess uncertainty at cn
	 * 
	 * @param cn
	 *            current node
	 * @param targetPrecision
	 *            goal width
	 * @param depth
	 *            current depth
	 * @param r
	 *            return value containing aStar and node with maximum excess
	 * @throws Exception
	 */
	public void getMaxExcess(Node cn, double targetPrecision, int depth,
			ArrayList r) throws Exception {
		// Action aStar is obtained from update upperbound.
		// We should find the appropriate observation oStar.
		int aStar = (Integer) r.get(0);

		if (Loggers.logSymHsviDdTreeFiniteSymmetry.isDebugEnabled()) {
			Loggers.logSymHsviDdTreeFiniteSymmetry.debug("    a=" + aStar);
		}

		ArrayList<Object> result = _getMaxExcess(cn.Q[aStar].outcomes,
				targetPrecision, depth);
		EdgeDd maxExcessEdgeDd = (EdgeDd) result.get(1);

		r.add(((EdgeDdLeaf) maxExcessEdgeDd).getEdge().nextNode);
	}

	/**
	 * get the next node with the maximum excess
	 * 
	 * @param dd
	 * @param targetPrecision
	 * @param depth
	 * @return ArrayList of (maximum excess value, EdgeDdLeaf with maximum
	 *         excess)
	 */
	private ArrayList _getMaxExcess(EdgeDd dd, double targetPrecision, int depth) {
		// it's a leaf
		double excess = Double.NEGATIVE_INFINITY;
		EdgeDd maxExcessEdgeDd = null;

		if (dd.getVar() == 0) {
			Edge childEdge = dd.getEdge();
			Node childNode = childEdge.nextNode;

			// if the current belief is not reachable (obsProb=0),
			// excess is negative infinity.
			if (childEdge.obsProb == 0) {
				excess = Double.NEGATIVE_INFINITY;
			} else {
				excess = childEdge.obsProb
						* ((childNode.ubVal - childNode.lbVal) - targetPrecision
								* Math.pow(p.discFact, -(depth + 1)));

				if (Loggers.logSymHsviDdTreeFiniteSymmetry.isDebugEnabled()) {
					Loggers.logSymHsviDdTreeFiniteSymmetry.debug("    obsProb="
							+ childEdge.obsProb + " nslb=" + childNode.lbVal
							+ " nsub=" + childNode.ubVal + " nsdiff="
							+ (childNode.ubVal - childNode.lbVal) + " width="
							+ excess);
				}
			}
			maxExcessEdgeDd = dd;
		} else {
			EdgeDd[] children = dd.getChildren();
			for (int i = 0; i < children.length; i++) {
				ArrayList<Object> resultBelow = _getMaxExcess(children[i],
						targetPrecision, depth);
				double maxExcess = (Double) resultBelow.get(0);
				if (excess < maxExcess) {
					excess = maxExcess;
					maxExcessEdgeDd = (EdgeDd) resultBelow.get(1);
				}
			}
		}

		ArrayList<Object> result = new ArrayList<Object>();
		result.add(excess);
		result.add(maxExcessEdgeDd);
		return result;
	}

	/**
	 * Pruning upper and lower bounds
	 * 
	 * @param horizon
	 *            pruning horizon
	 */
	private void prune(int horizon) {
		lowerBound.mayPrune(horizon);
		upperBound.mayPrune(horizon);
	}

	/**
	 * Select the depth+1 child that most contributes to excess uncertainty at
	 * cn.
	 * 
	 * @param cn
	 *            current node
	 * @param targetPrecision
	 *            goal width
	 * @param depth
	 *            current depth
	 * @param r
	 *            Return values containing aStar and oStar
	 * @throws Exception
	 */
	private void update(Node cn, double targetPrecision, int depth, ArrayList r)
			throws Exception {
		lowerBound.update(cn);

		// update upperbound and get the action 'aStar' in r.
		upperBound.update(cn, r);
		prune(cn.horizon);
		getMaxExcess(cn, targetPrecision, depth, r);
	}

	/**
	 * Initialize lowerBound and upperBound
	 */
	private void initBound() {
		double startTime = System.currentTimeMillis();
		// for approximating large problems, only initialize to initHorizon
		lowerBound.initBound(initHorizon);
		upperBound.initBound(initHorizon);
		double elapsedTime = System.currentTimeMillis() - startTime;
		Loggers.logSymHsviDdTreeFiniteSymmetry.info("elapsed time = "
				+ (elapsedTime / 1000.0) + " s");
	}

	/**
	 * Expand the POMDP tree structure at node cn
	 * 
	 * @param cn
	 *            expanding node
	 */
	public void expand(Node cn) {
		DD[] list1 = new DD[p.actions[0].transFn.length + p.nStateVars];
		DD[] list2 = new DD[p.actions[0].obsFn.length + 1];

		// If cn is new node
		if (cn.Q == null) {
			// Make QEntry for each action
			cn.Q = new QEntry[p.nActions];

			for (int a = 0; a < p.nActions; a++) {
				QEntry Qa = new QEntry();
				Qa.immediateReward = OP.factoredExpectationSparseNoMem(cn.b,
						p.actions[a].rewFn);

				for (int i = 0; i < p.nStateVars; i++)
					list1[i] = cn.b[i];

				for (int i = 0; i < p.actions[a].transFn.length; i++)
					list1[i + p.nStateVars] = p.actions[a].transFn[i];

				// sum_s b(s)T(s,a,s')
				DD temp = OP.addMultVarElim(list1, p.stateVarsId);
				list2[0] = temp;
				for (int i = 0; i < p.actions[a].obsFn.length; i++)
					list2[i + 1] = p.actions[a].obsFn[i];

				// dd_obsProbs(s',z) = O(s',a,z) sum_s T(s,a,s')b(s)
				DD dd_obsProbs = OP.multN(list2);

				// obsProbs(z) = sum_s' O(s',a,z) sum_s T(s,a,s')b(s)
				DD obsProbs = OP
						.addMultVarElim(dd_obsProbs, p.stateVarsPrimeId);

				Qa.outcomes = Search.getEdgeDd(cn.horizon, p, this,
						dd_obsProbs, obsProbs);

				Qa.outcomesDd = dd_obsProbs;
				Qa.obsProbs = obsProbs;
				Qa.ubVal = Double.NEGATIVE_INFINITY;
				cn.Q[a] = Qa;
			}
		}
	}

	/**
	 * Find and return the node which has b and primedB as belief state from the
	 * nodeBunch hash. If there is no corresponding node, make and return new
	 * node and input that node to nodeBunch has.
	 * 
	 * @param origDD
	 * @return returned node
	 */
	public ArrayList bringNode(int horizon, DD origDD) {

		DD[] b = new DD[p.nStateVars];
		for (int varId = 0; varId < p.nStateVars; varId++) {
			int[] vars = Util.elimOneVar(p.stateVarsId, varId);
			vars = MySet.remove(vars, 0);
			b[varId] = OP.addMultVarElim(origDD, vars);
		}

		// sort factored belief (state) vars
		ArrayList bPerm = sortStVars(b);
		b = (DD[]) bPerm.get(0);
		int[][] invPerm = (int[][]) bPerm.get(1);

		// If node with vector b already exists, bring the node
		Node dummyNode = new Node(b, p.stateVarsId, horizon);
		Node newNode = nodeBunch.get(dummyNode);

		if (newNode != null) {
			nodeHits++;
			ArrayList res = new ArrayList();
			res.add(newNode);
			for (int i = 0; i < 2; i++)
				for (int j = 0; j < p.nStateVars; j++)
					invPerm[i][j] += p.nVars;
			res.add(invPerm);
			return res;
		}

		// Else create new node
		newNode = dummyNode;
		lowerBound.initNodeBound(newNode);
		upperBound.initNodeBound(newNode);

		// Put newNode to nodeBunch
		nodeBunch.put(newNode, newNode);

		ArrayList res = new ArrayList();
		res.add(newNode);
		for (int i = 0; i < 2; i++)
			for (int j = 0; j < p.nStateVars; j++)
				invPerm[i][j] += p.nVars;
		res.add(invPerm);
		return res;
	}

	/**
	 * When use the symmetry, sort the variable
	 * 
	 * @param b
	 *            factored belief
	 * @return sorted factored belief and the permutation
	 */
	public ArrayList sortStVars(DD[] b) {
		if (varPerms.length == 1) {
			ArrayList res = new ArrayList();
			res.add(b);
			int[][] varMapping = new int[2][];
			varMapping[0] = new int[p.nStateVars];
			varMapping[1] = new int[p.nStateVars];
			for (int i = 0; i < p.nStateVars; i++) {
				varMapping[0][i] = i + 1;
				varMapping[1][i] = i + 1;
			}
			res.add(varMapping);
			return res;
		}

		// sort factored belief (state) vars
		double[] belVal = new double[p.nStateVars];
		for (int i = 0; i < p.nStateVars; i++) {
			belVal[i] = b[i].getChildren()[0].getVal();
		}

		ArrayList<int[]> permList = new ArrayList<int[]>();
		for (int pe = 0; pe < varPerms.length; pe++) {
			int[] newPerm = new int[p.nStateVars];
			for (int i = 0; i < p.nStateVars; i++)
				newPerm[i] = varPerms[pe][i];
			permList.add(newPerm);
		}
		for (int stVar = 0; stVar < p.nStateVars; stVar++) {
			ArrayList<int[]> newPermList = new ArrayList<int[]>();
			double maxVal = Double.NEGATIVE_INFINITY;
			for (int pe = 0; pe < permList.size(); pe++) {
				int[] perm = permList.get(pe);
				if (maxVal < belVal[perm[stVar] - 1]) {
					maxVal = belVal[perm[stVar] - 1];
					newPermList.clear();
					newPermList.add(perm);
				} else if (maxVal == belVal[perm[stVar] - 1]) {
					newPermList.add(perm);
				}
			}
			permList = newPermList;
			if (permList.size() == 1)
				break;
		}

		int[] perm = permList.get(0);
		int[][] varMapping = new int[2][];
		varMapping[0] = perm;
		varMapping[1] = new int[p.nStateVars];

		for (int i = 0; i < p.nStateVars; i++)
			varMapping[1][i] = i + 1;
		DD[] newB = new DD[p.nStateVars];
		for (int i = 0; i < p.nStateVars; i++) {
			newB[i] = OP.swapVars(b[perm[i] - 1], varMapping);
		}

		ArrayList res = new ArrayList();
		res.add(newB);
		int[][] invVarMapping = new int[2][];
		invVarMapping[0] = varMapping[1].clone();
		invVarMapping[1] = varMapping[0].clone();
		res.add(invVarMapping);
		return res;
	}

	/**
	 * Get alpha vectors and policy
	 * 
	 * @return ArrayList containing vectors of alphas and policy
	 */
	public ArrayList getAlphas() {
		java.util.Vector<Alpha>[] alphas = lowerBound.getAlphas();
		ArrayList result = new ArrayList(2);
		java.util.Vector<DD>[] alphaVectors = new java.util.Vector[horizon];
		java.util.Vector<Integer>[] policy = new java.util.Vector[horizon];

		// use masking
		java.util.Vector<DD[]>[] mask = new java.util.Vector[horizon];

		for (int h = 0; h < horizon; h++) {
			alphaVectors[h] = new java.util.Vector<DD>();
			policy[h] = new java.util.Vector<Integer>();
			if (lowerBound.useMasking)
				mask[h] = new java.util.Vector<DD[]>();

			for (int i = 0; i < alphas[h].size(); i++) {
				alphaVectors[h].add(alphas[h].get(i).v);
				policy[h].add(alphas[h].get(i).a);

				if (lowerBound.useMasking)
					mask[h].add(alphas[h].get(i).mask);
			}
		}

		result.add(alphaVectors);
		result.add(policy);
		if (lowerBound.useMasking)
			result.add(mask);

		return result;
	}

	/**
	 * read permutations
	 * 
	 * @param filename
	 *            permutation file
	 * @return permutation tables for state, obs variable & action
	 */
	public ArrayList readPermutations(String filename) {
		int[][] varPerms = null;
		int[][] actPerms = null;

		try {
			System.out.println("Read permutations from " + filename);
			BufferedReader br = new BufferedReader(new FileReader(filename));
			int nPerms = Integer.parseInt(br.readLine());
			System.out.println("# permutations = " + nPerms);
			varPerms = new int[nPerms][p.nVars];
			actPerms = new int[nPerms][p.nActions];

			for (int i = 0; i < nPerms; i++) {
				String perm = br.readLine();
				StringTokenizer st = new StringTokenizer(perm, " ");
				for (int j = 0; j < p.nVars; j++)
					varPerms[i][j] = Integer.parseInt(st.nextToken());
				for (int j = 0; j < p.nActions; j++)
					actPerms[i][j] = Integer.parseInt(st.nextToken());
			}
			br.close();
		} catch (Exception e) {
			System.out.println("no permutation info.");
			varPerms = new int[1][p.nVars];
			actPerms = new int[1][p.nActions];
			for (int v = 0; v < p.nVars; v++)
				varPerms[0][v] = v + 1;
			for (int ac = 0; ac < p.nActions; ac++)
				actPerms[0][ac] = ac;
		}

		boolean bPrint = true;
		if (bPrint) {
			System.out.println(" * Permuations");
			for (int i = 0; i < varPerms.length; i++) {
				for (int j = 0; j < p.nVars; j++) {
					if (j == 0)
						System.out.print("StVar: ");
					if (j == p.nStateVars)
						System.out.print(" ObsVar: ");
					System.out.printf("%3d ", varPerms[i][j]);
				}
				System.out.print(" Act: ");
				for (int j = 0; j < p.nActions; j++)
					System.out.printf("%3d ", actPerms[i][j]);
				System.out.println();
			}
		}

		ArrayList retList = new ArrayList();
		retList.add(varPerms);
		retList.add(actPerms);
		return retList;
	}
}