package kr.ac.kaist.pomdp.symbolic.perseus;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Vector;

import kr.ac.kaist.pomdp.utils.Util;
import no.uib.cipr.matrix.Matrix;

/**
 * SymbolicProblem class
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SymbolicProblem implements Serializable
{
	private static final long serialVersionUID = -2421781919892432972L;

	public ParseSpudd p;
	public int nStateVars;
	public int nObsVars;
	public int nVars;
	public int nActions;
	public double discFact;
	public double tolerance;
	public StateVarStruct[] stateVars;
	public ObsVarStruct[] obsVars;
	public ActStruct[] actions;
	public DD initialBelState;

	public int[] stateVarsId;
	public int[] stateVarsPrimeId;
	public int[] obsVarsId;
	public int[] obsVarsPrimeId;
	public int[] obsVarsArity;
	public int nObservations;

	public int[] stateObsVarsIds;
	public int[] stateVarsIds;
	public int[][] obsConfig;
	public int[][] tmpConfig;

	public int horizon;
	public String rddlInstanceName;

	private int[][] varPerms;
	private int[][] actPerms;
	
	public SymbolicProblem()
	{
	}

	/**
	 * SymbolicProblem constructor
	 * 
	 * @param filename
	 *            POMDP specification filename
	 * @param fullyObservable
	 */
	public SymbolicProblem(String filename, boolean fullyObservable)
	{

		System.out.println("Parsing " + filename);
		p = new ParseSpudd(filename);
		p.parsePOMDP(fullyObservable);
		dd_pomdp_flatAct(p);
		//		 display();

		// Initialize IDs
		stateVarsId = new int[p.nStateVars];
		stateVarsPrimeId = new int[p.nStateVars];
		obsVarsId = new int[p.nObsVars];
		obsVarsPrimeId = new int[p.nObsVars];
		obsVarsArity = new int[p.nObsVars];

		for (int varId = 0; varId < p.nStateVars; varId++)
		{
			stateVarsId[varId] = stateVars[varId].id;
			stateVarsPrimeId[varId] = stateVars[varId].id + nVars;
		}
		int nTemp = 1;
		for (int varId = 0; varId < p.nObsVars; varId++)
		{
			obsVarsId[varId] = obsVars[varId].id;
			obsVarsPrimeId[varId] = obsVars[varId].id + nVars;
			obsVarsArity[varId] = obsVars[varId].arity;
			nTemp *= obsVarsArity[varId];
		}
		nObservations = nTemp;

		stateObsVarsIds = Util.concat(stateVarsPrimeId, obsVarsPrimeId);
		stateVarsIds = Util.concat(stateVarsId, stateVarsPrimeId);

		obsConfig = new int[nObservations][];
		for (int obsId = 0; obsId < nObservations; obsId++)
		{
			obsConfig[obsId] = statedecode(obsId + 1, nObsVars, obsVarsArity);
		}
		tmpConfig = new int[2][];

		if(p.horizon != null)
			this.horizon = (int) p.horizon.getVal();
		else
			this.horizon = 100;
		this.rddlInstanceName = "";

		this.varPerms = null;
		this.actPerms = null;
		
		clearVars();
	}

	public void setPerms(int[][] _varPerms, int[][] _actPerms)
	{
		this.varPerms = _varPerms;
		this.actPerms = _actPerms;
	}
	
	private void clearVars()
	{
		p = null;
	}

	/**
	 * construct this SymbolicProblem from ParseSPUDD (called in SymbolicProblem
	 * constructor)
	 * 
	 * @param p
	 *            input ParseSPUDD instance
	 */
	public void dd_pomdp_flatAct(ParseSpudd p)
	{

		this.nStateVars = p.nStateVars;
		this.nObsVars = p.nObsVars;
		this.nVars = nStateVars + nObsVars;

		int[] newVarDomSize = new int[nStateVars * 2 + nObsVars * 2];
		String varNames[] = new String[nStateVars * 2 + nObsVars * 2];
		int i_newVar = 0;

		// State variables
		stateVars = new StateVarStruct[nStateVars];

		for (int varId = 0; varId < nStateVars; varId++)
		{
			stateVars[varId] = new StateVarStruct();
			stateVars[varId].arity = p.valNames.get(varId).size();
			stateVars[varId].name = p.varNames.get(varId);
			stateVars[varId].id = varId + 1;
			stateVars[varId].valNames = new String[stateVars[varId].arity];
			for (int valId = 0; valId < stateVars[varId].arity; valId++)
			{
				stateVars[varId].valNames[valId] = (String) p.valNames.get(varId).get(valId);
			}

			newVarDomSize[i_newVar] = stateVars[varId].arity;
			varNames[i_newVar] = stateVars[varId].name;
			i_newVar++;
		}

		// Observation variables
		obsVars = new ObsVarStruct[nObsVars];
		for (int varId = 0; varId < nObsVars; varId++)
		{
			obsVars[varId] = new ObsVarStruct();
			obsVars[varId].arity = p.valNames.get(nStateVars + varId).size();
			obsVars[varId].name = p.varNames.get(nStateVars + varId);
			obsVars[varId].id = nStateVars + varId + 1;
			obsVars[varId].valNames = new String[obsVars[varId].arity];
			for (int valId = 0; valId < obsVars[varId].arity; valId++)
			{
				obsVars[varId].valNames[valId] =
						(String) p.valNames.get(nStateVars + varId).get(valId);
			}
			newVarDomSize[i_newVar] = obsVars[varId].arity;
			varNames[i_newVar] = obsVars[varId].name;
			i_newVar++;
		}

		// Globals
		for (int i = 0; i < nVars; i++)
		{
			newVarDomSize[nVars + i] = newVarDomSize[i];
			varNames[nVars + i] = varNames[i] + "_P";
		}
		Global.setVarDomSize(newVarDomSize);
		Global.setVarNames(varNames);

		for (int i = 0; i < nStateVars; i++)
		{
			Global.setValNames(i + 1, stateVars[i].valNames);
		}
		for (int i = 0; i < nObsVars; i++)
		{
			Global.setValNames(nStateVars + i + 1, obsVars[i].valNames);
		}
		for (int i = 0; i < nStateVars; i++)
		{
			Global.setValNames(nVars + i + 1, stateVars[i].valNames);
		}
		for (int i = 0; i < nObsVars; i++)
		{
			Global.setValNames(nVars + nStateVars + i + 1, obsVars[i].valNames);
		}

		// Dynamics
		nActions = p.actTransitions.size();
		actions = new ActStruct[nActions];
		int[] vars = new int[nStateVars];

		for (int i = 0; i < nStateVars; i++)
		{
			vars[i] = nVars + i + 1;
			// System.out.println(stateVars[i].id);

		}
		for (int actId = 0; actId < nActions; actId++)
		{
			actions[actId] = new ActStruct();
			actions[actId].name = p.actNames.get(actId);
			actions[actId].transFn = p.actTransitions.get(actId);

			actions[actId].obsFn = p.actObserve.get(actId);
			actions[actId].rewFn = OP.sub(p.reward, p.actCosts.get(actId));

			// actions[actId].rewFn.display();

			DD[] dds = new DD[actions[actId].transFn.length + 1];
			dds[0] = actions[actId].rewFn;

			for (int trId = 0; trId < actions[actId].transFn.length; trId++)
			{
				dds[trId + 1] = actions[actId].transFn[trId];
				// actions[actId].transFn[trId].display();
			}
			// System.out.println("actId"+actId+" "+p.actNames.get(actId));
			actions[actId].rewFn = OP.addMultVarElim(dds, vars);

			// actions[actId].rewFn.display();
		}

		// Discount Factor
		discFact = p.discount.getVal();

		// Tolerance
		double maxVal = 0;
		double minVal = 0;
		for (int actId = 0; actId < nActions; actId++)
		{
			maxVal = OP.maxAll(actions[actId].rewFn);
			minVal = OP.minAll(actions[actId].rewFn);

		}
		double maxDiffRew = maxVal - minVal;
		double maxDiffVal = maxDiffRew / (1.0 - Math.min(0.95, discFact));
		tolerance = 0.00001 * maxDiffVal;

		double[] preVal = { 0.0 };

		for (int actId = 0; actId < nActions; actId++)
		{
			actions[actId].rewFn = OP.approximate(actions[actId].rewFn, tolerance, preVal);
		}

		// Init
		initialBelState = p.init;
	}

	/**
	 * display information about this SymbolicProblem
	 */
	public void display()
	{
		System.out.println("nStateVars: " + nStateVars);
		System.out.println("nObsVars: " + nObsVars);
		System.out.println("nVars: " + nVars + "\n");

		System.out.println("stateVarsId : ");
		for(int i=0; i<stateVarsId.length; i++) {
			System.out.print(stateVarsId[i] + " ");
		}
		System.out.println();
		System.out.println("stateVarsPrimeId : ");
		for(int i=0; i<stateVarsPrimeId.length; i++) {
			System.out.print(stateVarsPrimeId[i] + " ");
		}
		System.out.println();
		System.out.println("obsVarsId : ");
		for(int i=0; i<obsVarsId.length; i++) {
			System.out.print(obsVarsId[i] + " ");
		}
		System.out.println();
		System.out.println("obsVarsPrimeId : ");
		for(int i=0; i<obsVarsPrimeId.length; i++) {
			System.out.print(obsVarsPrimeId[i] + " ");
		}
		System.out.println();
		
		for (int i = 0; i < stateVars.length; i++)
		{
			System.out.println("id: " + stateVars[i].id);
			System.out.println("arity: " + stateVars[i].arity);
			System.out.println("name: " + stateVars[i].name);
			System.out.println("values: ");
			for (int j = 0; j < stateVars[i].valNames.length; j++)
			{
				System.out.print(stateVars[i].valNames[j] + " ");
			}
			System.out.println("\n");
		}
		System.out.println("obsVars: " + obsVars.length);
		for (int i = 0; i < obsVars.length; i++)
		{
			System.out.println("id: " + obsVars[i].id);
			System.out.println("arity: " + obsVars[i].arity);
			System.out.println("name: " + obsVars[i].name);
			System.out.println("values: ");
			for (int j = 0; j < obsVars[i].valNames.length; j++)
			{
				System.out.print(obsVars[i].valNames[j] + " ");
			}
			System.out.println("\n");
		}
		
		System.out.println("nActions: " + nActions);
		System.out.println("actions: " + actions.length + "\n");
		for (int i = 0; i < actions.length; i++)
		{
			System.out.println(actions[i].name);
			System.out.println("Transition:");
			for (int j = 0; j < actions[i].transFn.length; j++)
			{
				actions[i].transFn[j].display();
			}
			System.out.println("Reward:");
			actions[i].rewFn.display();
			System.out.println("Observation:");
			for (int j = 0; j < actions[i].obsFn.length; j++)
			{
				actions[i].obsFn[j].display();
			}
		}
		System.out.println("init belief: ");
		initialBelState.printSpuddDD(System.out);
		System.out.println();
		System.out.println("discFact: " + discFact);
		System.out.println("horizon: " + horizon);
		System.out.println("tolerance: " + tolerance);
		System.out.println();
	}

	/**
	 * get factored belief
	 * @param b
	 * @return
	 */
	public DD[] factorBelief(DD b)
	{
		DD[] factorb = new DD[nStateVars];
		for (int varId = 0; varId < nStateVars; varId++)
		{
			int[] vars = Util.elimOneVar(stateVarsId, varId);
			vars = MySet.remove(vars, 0);
			factorb[varId] = OP.addMultVarElim(b, vars);
		}

		return factorb;
	}

	/**
	 * simulation
	 * 
	 * @param alphas
	 *            alpha vectors
	 * @param policy
	 * @param masks
	 * @param nRuns
	 *            the number of simulation runs
	 * @param nSteps
	 *            the maximum number of steps
	 * @param seed
	 *            random seed number
	 * @param debug
	 *            true or false
	 * @return result matrix (run number by 3) column 0: reward column 1: step
	 *         length column 2: 1 if this run was a success, 0 otherwise.
	 */
	public Matrix sampleRewards(Vector<DD> alphas, Vector<Integer> policy, Vector<DD[]> masks,
			int nRuns, int nSteps, int seed, boolean debug)
	{
		double rew = 0.0;

		// result matrix
		no.uib.cipr.matrix.Matrix result = new no.uib.cipr.matrix.DenseMatrix(nRuns, 3);
		Global.random.setSeed(seed);

		// for each run,
		for (int run = 0; run < nRuns; run++)
		{

			if (debug)
				System.out.println("runId " + run);

			// initial belief state
			DD belState = initialBelState.store();

			// sample one state configuration
			int[][] stateConfig = OP.sampleMultinomial(belState, stateVarsId);

			double totRew = 0.0; // total reward for this run
			int runStep = 0;
			int success = 0;
			for (int stepId = 0; stepId < nSteps; stepId++)
			{

				// printBelief(belState);

				// get the optimal action and reward
				int actId = policyQueryNew(belState, alphas, policy, masks);
				rew = OP.eval(actions[actId].rewFn, stateConfig);
				//				totRew = totRew + rew;
				totRew = totRew + Math.pow(this.discFact, stepId) * rew;

				// sample the state and observation of the next step
				DD[] restrictedTransFn = OP.restrictN(actions[actId].transFn, stateConfig);
				int[][] stateConfig2 = OP.sampleMultinomial(restrictedTransFn, stateVarsPrimeId);
				DD[] restrictedObsFn =
						OP.restrictN(actions[actId].obsFn, Util.concat(stateConfig, stateConfig2));
				int[][] obsConfig = OP.sampleMultinomial(restrictedObsFn, obsVarsPrimeId);

				// belief update
				belState = beliefUpdate(belState, actId, obsConfig);
				stateConfig = Config.primeVars(stateConfig2, -nVars);

				// for debug
				if (debug)
				{
					System.out.println((stepId + 1) + " step, action " + actions[actId].name
							+ ", rew = " + rew);
					for (int k = 0; k < nObsVars; k++)
						System.out.println(obsVars[k].name + ": "
								+ obsVars[k].valNames[obsConfig[1][k] - 1]);
					for (int k = 0; k < nStateVars; k++)
						System.out.println(stateVars[k].name + ": "
								+ stateVars[k].valNames[stateConfig[1][k] - 1]);
				}

				Global.newHashtables();
			}

			if (runStep == 0)
				runStep = nSteps;
			result.set(run, 0, totRew);
			result.set(run, 1, runStep);
			result.set(run, 2, success);

			// total reward for this run >=0, this run is a success.
			if (totRew >= 0)
				success++;
		}
		return result;
	}

	/**
	 * simulation
	 * 
	 * @param alphas
	 *            alpha vectors
	 * @param policy
	 *            input policy
	 * @param nRuns
	 *            the number of simulation runs
	 * @param nSteps
	 *            the maximum number of steps
	 * @param seed
	 *            random seed number
	 * @param debug
	 *            true or false
	 * @return result matrix (run number by 3) column 0: reward column 1: step
	 *         length column 2: 1 if this run was a success, 0 otherwise.
	 */
	public no.uib.cipr.matrix.Matrix sampleRewards(Vector alphas, Vector policy, int nRuns,
			int nSteps, int seed, boolean debug)
	{

		double rew = 0.0;

		// result matrix
		no.uib.cipr.matrix.Matrix result = new no.uib.cipr.matrix.DenseMatrix(nRuns, 3);
		Global.random.setSeed(seed);

		// for each run,
		for (int run = 0; run < nRuns; run++)
		{

			if (debug)
				System.out.println("runId " + run);

			// initial belief state
			DD belState = initialBelState.store();

			// sample one state configuration
			int[][] stateConfig = OP.sampleMultinomial(belState, stateVarsId);

			double totRew = 0.0; // total reward for this run
			int runStep = 0;
			int success = 0;
			for (int stepId = 0; stepId < nSteps; stepId++)
			{

//				 printBelief(belState);

				// get the optimal action and reward
				int actId = policyQuery(belState, alphas, policy);
				rew = OP.eval(actions[actId].rewFn, stateConfig);
				//				totRew = totRew + rew;
				totRew = totRew + Math.pow(this.discFact, stepId) * rew;

				// sample the state and observation of the next step
				DD[] restrictedTransFn = OP.restrictN(actions[actId].transFn, stateConfig);
				int[][] stateConfig2 = OP.sampleMultinomial(restrictedTransFn, stateVarsPrimeId);
				DD[] restrictedObsFn =
						OP.restrictN(actions[actId].obsFn, Util.concat(stateConfig, stateConfig2));
				int[][] obsConfig = OP.sampleMultinomial(restrictedObsFn, obsVarsPrimeId);

				//				if (Math.abs(rew - goalrew) < 1e-5) {
				//					runStep = stepId + 1;
				//					success = 1;
				//					break;
				//				}

				// belief update
				belState = beliefUpdate(belState, actId, obsConfig);
				stateConfig = Config.primeVars(stateConfig2, -nVars);

				// for debug
				if (debug)
				{
					System.out.println((stepId + 1) + " step, action " + actions[actId].name
							+ ", rew = " + rew);
					for (int k = 0; k < nObsVars; k++)
						System.out.println(obsVars[k].name + ": "
								+ obsVars[k].valNames[obsConfig[1][k] - 1]);
					for (int k = 0; k < nStateVars; k++)
						System.out.println(stateVars[k].name + ": "
								+ stateVars[k].valNames[stateConfig[1][k] - 1]);
				}

				Global.newHashtables();
			}

			if (runStep == 0)
				runStep = nSteps;
			result.set(run, 0, totRew);
			result.set(run, 1, runStep);
			result.set(run, 2, success);

			// total reward for this run >=0, this run is a success.
			if (totRew >= 0)
				success++;
		}
		return result;
	}

	/**
	 * simulation
	 * 
	 * @param alphas
	 *            alpha vectors
	 * @param policy
	 *            input policy
	 * @param nRuns
	 *            the number of simulation runs
	 * @param nSteps
	 *            the maximum number of steps
	 * @param seed
	 *            random seed number
	 * @param debug
	 *            true or false
	 * @return result matrix (run number by 3) column 0: reward column 1: step
	 *         length column 2: 1 if this run was a success, 0 otherwise.
	 */
	public no.uib.cipr.matrix.Matrix sampleRewardsFinite(Vector[] alphas, Vector[] policy,
			int nRuns, int nSteps, int seed, boolean debug)
	{

		double rew = 0.0;
		double goalrew = 10.0; // if the reward is same to goalrew, simulation
		// run will stop and be a success.

		// result matrix
		no.uib.cipr.matrix.Matrix result = new no.uib.cipr.matrix.DenseMatrix(nRuns, 3);
		Global.random.setSeed(seed);

		// for each run,
		for (int run = 0; run < nRuns; run++)
		{

			if (debug)
				System.out.println("runId " + run);

			// initial belief state
			DD belState = initialBelState.store();

			// sample one state configuration
			int[][] stateConfig = OP.sampleMultinomial(belState, stateVarsId);

			double totRew = 0.0; // total reward for this run
			int runStep = 0;
			int success = 0;
			for (int stepId = 0; stepId < nSteps; stepId++)
			{

				// printBelief(belState);

				// get the optimal action and reward
				int actId =
						policyQuery(belState, alphas[horizon - stepId - 1], policy[horizon - stepId
								- 1]);
				rew = OP.eval(actions[actId].rewFn, stateConfig);
				//				totRew = totRew + rew;
				totRew = totRew + Math.pow(this.discFact, stepId) * rew;

				// sample the state and observation of the next step
				DD[] restrictedTransFn = OP.restrictN(actions[actId].transFn, stateConfig);
				int[][] stateConfig2 = OP.sampleMultinomial(restrictedTransFn, stateVarsPrimeId);
				DD[] restrictedObsFn =
						OP.restrictN(actions[actId].obsFn, Util.concat(stateConfig, stateConfig2));
				int[][] obsConfig = OP.sampleMultinomial(restrictedObsFn, obsVarsPrimeId);

				if (Math.abs(rew - goalrew) < 1e-5)
				{
					runStep = stepId + 1;
					success = 1;
					break;
				}

				// belief update
				belState = beliefUpdate(belState, actId, obsConfig);
				stateConfig = Config.primeVars(stateConfig2, -nVars);

				// for debug
				if (debug)
				{
					System.out.println((stepId + 1) + " step, action " + actions[actId].name
							+ ", rew = " + rew);
					for (int k = 0; k < nObsVars; k++)
						System.out.println(obsVars[k].name + ": "
								+ obsVars[k].valNames[obsConfig[1][k] - 1]);
					for (int k = 0; k < nStateVars; k++)
						System.out.println(stateVars[k].name + ": "
								+ stateVars[k].valNames[stateConfig[1][k] - 1]);
				}

				Global.newHashtables();
			}

			if (runStep == 0)
				runStep = nSteps;
			result.set(run, 0, totRew);
			result.set(run, 1, runStep);
			result.set(run, 2, success);

			// total reward for this run >=0, this run is a success.
			if (totRew >= 0)
				success++;
		}
		return result;
	}
	
	public no.uib.cipr.matrix.Matrix sampleRewardsFiniteMask(Vector[] alphas, Vector[] policy, Vector[] mask, 
			int nRuns, int nSteps, int seed, boolean debug)
	{

		double rew = 0.0;
		double goalrew = 10.0; // if the reward is same to goalrew, simulation
		// run will stop and be a success.

		// result matrix
		no.uib.cipr.matrix.Matrix result = new no.uib.cipr.matrix.DenseMatrix(nRuns, 3);
		Global.random.setSeed(seed);

		// for each run,
		for (int run = 0; run < nRuns; run++)
		{

			if (debug)
				System.out.println("runId " + run);

			// initial belief state
			DD belState = initialBelState.store();

			// sample one state configuration
			int[][] stateConfig = OP.sampleMultinomial(belState, stateVarsId);

			double totRew = 0.0; // total reward for this run
			int runStep = 0;
			int success = 0;
			for (int stepId = 0; stepId < nSteps; stepId++)
			{
//System.out.println("step = " + stepId);
				// printBelief(belState);

				// get the optimal action and reward
				int actId =
						policyQueryNew(belState, alphas[horizon - stepId - 1], policy[horizon - stepId
								- 1], mask[horizon - stepId - 1]);
				rew = OP.eval(actions[actId].rewFn, stateConfig);
				//				totRew = totRew + rew;
				totRew = totRew + Math.pow(this.discFact, stepId) * rew;

				// sample the state and observation of the next step
				DD[] restrictedTransFn = OP.restrictN(actions[actId].transFn, stateConfig);
				int[][] stateConfig2 = OP.sampleMultinomial(restrictedTransFn, stateVarsPrimeId);
				DD[] restrictedObsFn =
						OP.restrictN(actions[actId].obsFn, Util.concat(stateConfig, stateConfig2));
				int[][] obsConfig = OP.sampleMultinomial(restrictedObsFn, obsVarsPrimeId);

				if (Math.abs(rew - goalrew) < 1e-5)
				{
					runStep = stepId + 1;
					success = 1;
					break;
				}

				// belief update
				belState = beliefUpdate(belState, actId, obsConfig);
				stateConfig = Config.primeVars(stateConfig2, -nVars);

				// for debug
				if (debug)
				{
					System.out.println((stepId + 1) + " step, action " + actions[actId].name
							+ ", rew = " + rew);
					for (int k = 0; k < nObsVars; k++)
						System.out.println(obsVars[k].name + ": "
								+ obsVars[k].valNames[obsConfig[1][k] - 1]);
					for (int k = 0; k < nStateVars; k++)
						System.out.println(stateVars[k].name + ": "
								+ stateVars[k].valNames[stateConfig[1][k] - 1]);
				}

				Global.newHashtables();
			}

			if (runStep == 0)
				runStep = nSteps;
			result.set(run, 0, totRew);
			result.set(run, 1, runStep);
			result.set(run, 2, success);

			// total reward for this run >=0, this run is a success.
			if (totRew >= 0)
				success++;
		}
		return result;
	}

	/**
	 * get the optimal action id on this stateConfig
	 * 
	 * @param mdpAlpha
	 *            MDP alpha vector
	 * @param stateConfig
	 *            current state configuration
	 * @param values
	 *            value function
	 * @return the optimal action id
	 */
	public int MDPpolicyQuery(DD mdpAlpha, int[][] stateConfig, DD[] values)
	{

		double[] rew_a = new double[nActions];

		for (int a = 0; a < nActions; a++)
		{
			rew_a[a] = OP.eval(values[a], stateConfig);
		}

		int result = Util.argmax(rew_a);
		return result;
	}

	/**
	 * get the optimal action id on this belief point
	 * 
	 * @param belState
	 *            current belief point
	 * @param alphaVectors
	 * @param policy
	 * @return the optimal action id
	 */
	public int policyQuery(DD belState, Vector alphaVectors, Vector policy)
	{
		int bestAlphaId = -1;

//		// factored belief state
//		DD[] b = factorBelief(belState);
//		ArrayList sortRes = sortStVars(b);
//		b = (DD[]) sortRes.get(0);
//		int numActPerm = (Integer) sortRes.get(1);
		
		double bestVal = Double.NEGATIVE_INFINITY;
		double val = 0.0;
		for (int alphaId = 0; alphaId < alphaVectors.size(); alphaId++)
		{

			val = OP.dotProduct(belState, (DD) alphaVectors.get(alphaId), stateVarsId);
//			val = OP.factoredExpectationSparseNoMem(b, (DD) alphaVectors.get(alphaId));

			if (val > bestVal)
			{
				bestVal = val;
				bestAlphaId = alphaId;
			}
		}
		
		int act = -1;
//		if(policy.size() > 0) 
			act = (Integer) policy.get(bestAlphaId);
//		else {
//			act = Global.random.nextInt(this.nActions);
//			System.out.println("                  !!!!!!!");
//			act = 0;
//		}
//		System.out.println("no supp act = " + this.actions[act].name);
		
//		return actPerms[numActPerm][act];
		return act;
	}

	/**
	 * get the optimal action id on this belief point
	 * 
	 * @param belState current belief point
	 * @param alphaVectors
	 * @return the optimal action id
	 */
	public int policyQueryNew(DD belState, Vector<DD> alphaVectors, Vector<Integer> policy,
			Vector<DD[]> masks)
	{
		int bestAlphaId = 0;

		// factored belief state
		DD[] b = factorBelief(belState);
		ArrayList sortRes = sortStVars(b);
		b = (DD[]) sortRes.get(0);
		int numActPerm = (Integer) sortRes.get(1);

		double bestVal = Double.NEGATIVE_INFINITY;
		double val = 0.0;
		for (int alphaId = 0; alphaId < alphaVectors.size(); alphaId++)
		{
			DD alpha = alphaVectors.get(alphaId);
			DD[] mask = masks.get(alphaId);

			// if the current belief b doesn't support alpha vector, just skip
			if (!OP.isSupport(b, mask))
				continue;

//			val = OP.dotProduct(belState, alpha, stateVarsId);
			val = OP.factoredExpectationSparseNoMem(b, alpha);

			if (val > bestVal)
			{
				bestVal = val;
				bestAlphaId = alphaId;
			}
			
//			System.out.println("alphaId = " + alphaId + ",  val = " + val + ", act = " + (Integer)policy.get(alphaId));
			
		}
		
		
		int act = 0;
		if(policy.size() > 0) 
			act = (Integer) policy.get(bestAlphaId);
		else {
			act = Global.random.nextInt(this.nActions);
//			System.out.println("                  !!!!!!!");
//			act = 0;
//			act = this.nActions-1;
		}
		
//		System.out.println("act = " + this.actions[act].name);
		return actPerms[numActPerm][act];
	}
	
	public ArrayList sortStVars(DD[] b) {
		if(varPerms.length == 1) {
			ArrayList res = new ArrayList();
			res.add(b);
			res.add(0);
			return res;
		}
		
		// sort factored belief (state) vars
		double[] belVal = new double[nStateVars];
		for (int i = 0; i < nStateVars; i++) {
			belVal[i] = b[i].getChildren()[0].getVal();
		}
		
		ArrayList<int[]> permList = new ArrayList<int[]> ();
		for(int pe=0; pe<varPerms.length; pe++) {
			permList.add(varPerms[pe]);
		}
		for(int stVar=0; stVar<nStateVars; stVar++) {
			ArrayList<int[]> newPermList = new ArrayList<int[]> ();
			double maxVal = Double.NEGATIVE_INFINITY;
			for(int pe=0; pe<permList.size(); pe++) {
				int[] perm = permList.get(pe);
				if(maxVal < belVal[perm[stVar]-1]) {
					maxVal = belVal[perm[stVar]-1];
					newPermList.clear();
					newPermList.add(perm);
				} else if(maxVal == belVal[perm[stVar]-1]) {
					newPermList.add(perm);
				}
			}
			permList = newPermList;
			if(permList.size() == 1)
				break;
		}
		
		int[] perm = permList.get(0);
		int[][] varMapping = new int[2][];
		varMapping[0] = perm;
		varMapping[1] = new int[nStateVars];
		for (int i = 0; i < nStateVars; i++)
			varMapping[1][i] = i+1;
		DD[] newB = new DD[nStateVars];
		for (int i = 0; i < nStateVars; i++) {
			newB[i] = OP.swapVars(b[perm[i]-1], varMapping);
		}
		
		ArrayList res = new ArrayList();
		res.add(newB);
		
		int numActPerm = 0;
		for(int pe=0; pe<varPerms.length; pe++) {
			boolean bSame = true;
			for(int stVar=0; stVar<nStateVars; stVar++) {
				if(varPerms[pe][stVar] != perm[stVar]) {
					bSame = false;
					break;
				}
			}
			if(bSame == true) {
				numActPerm = pe;
				break;
			}
		}
		res.add(numActPerm);
		
		return res;
	}

	/**
	 * update belief state given current belief state, action, and observation
	 * 
	 * @param belState
	 *            current belief state
	 * @param actId
	 *            action id
	 * @param obsConfig
	 *            observation
	 * @return updated belief state
	 */
	public DD beliefUpdate(DD belState, int actId, int[][] obsConfig)
	{

		DD[] TransFn = actions[actId].transFn;
		DD[] ObsFn = OP.restrictN(actions[actId].obsFn, obsConfig);

		DD[] mergeDD = OP.concat(TransFn, belState);
		mergeDD = OP.concat(mergeDD, ObsFn);

		DD nextBelState = OP.addMultVarElim(mergeDD, stateVarsId);
		nextBelState = OP.primeVars(nextBelState, -nVars);
		DD obsProb = OP.addMultVarElim(nextBelState, stateVarsId);

		if (obsProb.getVal() < 1e-8)
		{
			System.out
					.println("WARNING: Zero-probability observation, resetting belief state to a uniform distribution\n");
			nextBelState = DD.one;
		}

		nextBelState = OP.div(nextBelState, OP.addMultVarElim(nextBelState, stateVarsId));
		return nextBelState;
	}

	/**
	 * decode state id to factored configuration
	 * 
	 * @param statenum
	 *            input state id
	 * @param n
	 *            the number of state variables
	 * @param bases
	 *            the number of possible values for each state variables
	 * @return factored configuration
	 */
	public int[] statedecode(int statenum, int n, int[] bases)
	{
		int[] stateVec = new int[n];
		for (int i = 0; i < n; i++)
			stateVec[i] = 0;

		if (statenum == 1)
		{
			for (int i = 0; i < n; i++)
				stateVec[i]++;
			return stateVec;
		}
		statenum--;
		int res = statenum;
		for (int i = 0; i < n; i++)
		{
			if (res == 1)
			{
				stateVec[i] = 1;
				break;
			}
			int remd = res % bases[i];
			res = (int) Math.floor((double) res / bases[i]);
			stateVec[i] = remd;
		}
		for (int i = 0; i < n; i++)
			stateVec[i]++;

		return stateVec;
	}

	/**
	 * print belief state
	 * 
	 * @param belState
	 *            input belief state
	 */
	public void printBelief(DD belState)
	{

		// factorize initBelState
		DD[] temp = new DD[nStateVars];

		for (int varId = 0; varId < nStateVars; varId++)
		{
			int[] vars = Util.elimOneVar(stateVarsId, varId);
			vars = MySet.remove(vars, 0);
			temp[varId] = OP.addMultVarElim(belState, vars);
			temp[varId].display();
		}
	}
}
