package kr.ac.kaist.pomdp.utils;

import java.io.Serializable;
import java.util.List;
import java.util.Vector;

import kr.ac.kaist.pomdp.symbolic.perseus.DD;
import kr.ac.kaist.pomdp.symbolic.perseus.OP;
import kr.ac.kaist.pomdp.symbolic.perseus.SymbolicProblem;

/**
 * SupportList class for speeding up alpha vector masking
 * 
 * @author Kanghoon Lee (khlee@ai.kaist.ac.kr)
 * @version 2011.06.26
 * 
 */
public class SupportList implements Serializable {
	private static final long serialVersionUID = -8902435910849182881L;

	private int nStateVars;
	private int[] arity;
	private Vector[][] supportList;

	/**
	 * Constructor
	 * 
	 * @param p
	 *            symbolic problem instance
	 */
	public SupportList(SymbolicProblem p) {
		nStateVars = p.nStateVars;
		arity = new int[nStateVars];
		for (int i = 0; i < nStateVars; i++)
			arity[i] = p.stateVars[i].arity;

		supportList = new Vector[nStateVars][];
		for (int i = 0; i < nStateVars; i++) {
			supportList[i] = new Vector[arity[i]];
			for (int j = 0; j < arity[i]; j++) {
				supportList[i][j] = new Vector();
			}
		}
	}

	/**
	 * add alpha to the support list with respect to its mask DD.
	 * 
	 * @param alpha
	 * @param mask
	 */
	public void add(Object alpha, DD[] mask) {
		if (mask == null) {
			System.err.println("mask is null");
		}

		for (int i = 0; i < nStateVars; i++) {
			List<int[][]> nonzeroConfigs = OP.findAllNonzeroLeaf(mask[i]);
			if (nonzeroConfigs == null)
				return;

			for (int[][] config : nonzeroConfigs) {
				// if there is no variable, all the leaf is nonzero.
				if (config[0].length == 0) {
					for (int j = 0; j < arity[i]; j++) {
						supportList[i][j].add(alpha);
					}
				} else {
					supportList[i][config[1][0] - 1].add(alpha);
				}
			}
		}
	}

	/**
	 * get the supporting alpha vector list for belief b (finding one possible
	 * state and return its corresponding alpha vector list)
	 * 
	 * @param b
	 * @return
	 */
	public Vector getList(DD[] b) {
		Vector result = null;
		for (int i = 0; i < nStateVars; i++) {
			int[][] nonzeroConfig = OP.findNonzeroLeaf(b[i]);

			// if there is no nonzero leaf in b[i],
			// there is no possible leaf in the belief
			if (nonzeroConfig == null)
				return null;

			// assert there is only one variable in b[i]
			if (nonzeroConfig[0].length > 1) {
				System.err.println("there is more than one variable in b[i]");
				return null;
			}

			Vector nonzeroVector = null;
			// if there is no variable, all the leaf is nonzero.
			if (nonzeroConfig[0].length == 0) {
				nonzeroVector = supportList[i][0];
			} else {
				nonzeroVector = supportList[i][nonzeroConfig[1][0] - 1];
			}

			// intersect the current nonzeroVector with result
			if (result == null)
				result = nonzeroVector;
			else {
				result = Util.intersect(result, nonzeroVector);
			}
		}

		return result;
	}

	public void remove(Object alpha, DD[] mask) {
		if (mask == null) {
			System.err.println("mask is null");
		}

		for (int i = 0; i < nStateVars; i++) {
			List<int[][]> nonzeroConfigs = OP.findAllNonzeroLeaf(mask[i]);
			if (nonzeroConfigs == null)
				return;

			for (int[][] config : nonzeroConfigs) {
				// if there is no variable, all the leaf is nonzero.
				if (config[0].length == 0) {
					for (int j = 0; j < arity[i]; j++) {
						supportList[i][j].remove(alpha);
					}
				} else {
					supportList[i][config[1][0] - 1].remove(alpha);
				}
			}
		}
	}
}
